from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from numpyencoder import NumpyEncoder

from MEDiml.learning.ml_utils import (combine_rad_tables, finalize_rad_table,
                                        get_stratified_splits,
                                        intersect_var_tables)
from MEDiml.utils.get_full_rad_names import get_full_rad_names
from MEDiml.utils.json_utils import save_json


class FSR:
    def __init__(self, method: str = 'fda') -> None:
        """
        Feature set reduction class constructor.

        Args:
            method (str): Method of feature set reduction. Can be "FDA", "LASSO" or "mRMR".
        """
        self.method = method

    def __get_fda_corr_table(
            self, 
            variable_table: pd.DataFrame, 
            outcome_table_binary: pd.DataFrame,
            n_splits: int, 
            corr_type: str, 
            seed: int
        ) -> pd.DataFrame:
        """
        Calculates the correlation table of the FDA algorithm.

        Args:
            variable_table (pd.DataFrame): variable table to check for stability.
            outcome_table_binary (pd.DataFrame): outcome table with binary labels.
            n_splits (int): Number of splits in the FDA algorithm (Ex: 100).
            corr_type: String specifying the correlation type that we are investigating. 
                Must be either 'Pearson' or 'Spearman'.
            seed (int): Random generator seed.
        
        Returns:
            pd.DataFrame: Correlation table of the FDA algorithm. Rows are splits, columns are features.
        """
        # Setting the seed
        np.random.seed(seed)

        # Initialization
        row_names = []
        corr_table = pd.DataFrame()
        fraction_for_splits = 1/3
        number_of_splits = 1

        # For each split, we calculate the correlation table
        for s in range(n_splits):
            row_names.append("Split_{0:03}".format(s))

            # Keep only variables that are in both tables
            _, outcome_table_binary = intersect_var_tables(variable_table, outcome_table_binary)

            # Under-sample the outcome table to equalize the number of positive and negative outcomes
            #outcome_table_binary_balanced = under_sample(outcome_table_binary)

            # Get the patient teach split
            patients_teach_splits = get_stratified_splits(
                outcome_table_binary, 
                number_of_splits, 
                fraction_for_splits, 
                seed, 
                flag_by_cat=True
            )[0]

            # Creating a table with both the variables and the outcome with
            # only the patient teach splits, ranked for spearman and not for pearson
            if corr_type == 'Spearman':
                full_table = pd.concat([variable_table.loc[patients_teach_splits, :].rank(),
                                        outcome_table_binary.loc[patients_teach_splits,
                                        outcome_table_binary.columns.values[-1]]], axis=1)

            elif corr_type == 'Pearson':
                # Pearson is the base method used by numpy, so we dont have to do any
                # manipulations to the data  like with spearman.
                full_table = pd.concat([variable_table.loc[patients_teach_splits, :],
                                        outcome_table_binary.loc[patients_teach_splits,
                                        outcome_table_binary.columns.values[-1]]], axis=1)
            else:
                raise ValueError("Correlation type not recognized. Please use 'Pearson' or 'Spearman'")

            # calculate the whole correlation table for all variables.
            full_table = np.corrcoef(full_table, rowvar=False)[-1][:-1].reshape((1, -1))
            corr_table = corr_table.append(pd.DataFrame(full_table))

        # Add the metadata to the correlation table
        corr_table.columns = list(variable_table.columns.values)
        corr_table = corr_table.fillna(0)
        corr_table.index = row_names
        corr_table.Properties = {}
        corr_table._metadata += ['Properties']
        corr_table.Properties['description'] = variable_table.Properties['Description']
        corr_table.Properties['userData'] = variable_table.Properties['userData']

        return corr_table
    
    def __find_fda_best_mean(self, corr_tables: pd.DataFrame, min_n_feat_stable: int) -> Tuple[Dict, pd.DataFrame]:
        """
        Finds the best mean correlation of all the stable variables in the table.

        Args:
            corr_tables (Dict): dictionary containing the correlation tables of
                dimension : [n_splits,n_features] for each table.
            min_n_feat_stable (int): minimal number of stable features.

        Returns:
            Tuple[Dict, pd.DataFrame]: Dict containing the name of each stable variables in every table and 
                pd.DataFrame containing the mean correlation of all the stable variables in the table.
        """
        # Initialization
        var_names_stable = {}
        corr_mean_stable = corr_tables
        n_features = 0
        corr_table = corr_tables
        corr_table = corr_table.fillna(0)

        # Calculation of the mean correlation among the n splits (R mean)
        var_names_stable = corr_table.index

        # Calculating the total number of features
        n_features += var_names_stable.size

        # Getting absolute values of the mean correlation
        corr_mean_stable_abs = corr_mean_stable.abs()

        # Keeping only the best features if there are more than min_n_feat_stable features
        if n_features > min_n_feat_stable:
            # Get min_n_feat_stable highest correlations
            best_features = corr_mean_stable_abs.sort_values(ascending=False)[0:min_n_feat_stable]
            var_names_stable = best_features.index.values
            corr_mean_stable = best_features

        return var_names_stable, corr_mean_stable

    def __find_fda_stable(self, corr_table: pd.DataFrame, thresh_stable: float) -> Tuple[Dict, pd.DataFrame]:
        """
        Finds the stable features in each correlation table 
        and the mean correlation of all the stable variables in the table.
        
        Args:
            corr_tables (Dict): dictionary containing the correlation tables of
                dimension : [n_splits,n_features] for each table.
            thresh_stable (float): the threshold deciding if a feature is stable.
        
        Returns:
            Tuple[Dict, pd.DataFrame]: dictionary containing the name of each stable variables in every tables 
                and table containing the mean correlation of all the stable variables in the table. 
                (The keys are the table names and the values are pd.Series).
        """

        # Initialization
        corr_table.fillna(0, inplace=True)

        # Calculation of R mean
        corr_mean_stable = corr_table.mean()
        mean_r = corr_mean_stable

        # Calculation of min and max
        min_r = corr_table.quantile(0.05)
        max_r = corr_table.quantile(0.95)

        # Calculation of unstable features
        unstable = (min_r < thresh_stable) & (mean_r > 0) | (max_r > -thresh_stable) & (mean_r < 0)
        ind_unstable = unstable.index[unstable]

        # Stable variables
        var_names_stable = unstable.index[~unstable].values
        corr_mean_stable = mean_r.drop(ind_unstable)

        return var_names_stable, corr_mean_stable

    def __keep_best_text_param(
            self, 
            corr_table: pd.DataFrame, 
            var_names_stable: List, 
            corr_mean_stable: pd.DataFrame
        ) -> Tuple[List, pd.DataFrame]:
        """
        Keeps the best texture features extraction parameters in the correlation tables
        by dropping the variants of a given feature.

        Args:
            corr_table (pd.DataFrame): Correlation table of dimension : [n_splits,n_features].
            var_names_stable (List): List of the stable variables in the table.
            corr_mean_stable (pd.DataFrame): Table of the mean correlation of the stable variables in the variables table.
        
        Returns:
            Tuple[List, pd.DataFrame]: list of the stable variables in the tables and table containing the mean 
                correlation of all the stable variables.
        """

        # If no stable features for the currect field, continue
        if var_names_stable.size == 0:
            return var_names_stable, corr_mean_stable

        # Get the actual radiomics features names from the sequential names
        full_rad_names = get_full_rad_names( 
            corr_table.Properties['userData']['variables']['var_def'], 
            var_names_stable)

        # Now parsing the full names to get only the rad names and not the variant
        rad_names = np.array([])
        for n in range(full_rad_names.size):
            rad_names = np.append(rad_names, full_rad_names[n].split('__')[1:2])

        # Verifying if two features are the same variant and keeping the best one
        n_var = rad_names.size
        var_to_drop = []
        for rad_name in rad_names:
            # If all the features are unique, break
            if np.unique(rad_names).size == n_var:
                break
            else:
                ind_same = np.where(rad_names == rad_name)[0]
                n_same = ind_same.size
                if n_same > 1:
                    var_to_drop.append(list(corr_mean_stable.iloc[ind_same].sort_values().index[1:].values))
        
        # Dropping the variants
        if len(var_to_drop) > 0:
            # convert to list of lists to list
            var_to_drop = [item for sublist in var_to_drop for item in sublist]
            
            # From the unique values of var_to_drop, drop the variants
            for var in set(var_to_drop):
                var_names_stable = np.delete(var_names_stable, np.where(var_names_stable == var))
                corr_mean_stable = corr_mean_stable.drop(var)

        return var_names_stable, corr_mean_stable

    def __remove_correlated_variables(
            self, 
            variable_table: pd.DataFrame, 
            rank: pd.Series,
            corr_type: str, 
            thresh_inter_corr: float,
            min_n_feat_total: int
        ) -> pd.DataFrame:
        """
        Removes inter-correlated variables given a certain threshold.
        
        Args:
            variable_table (pd.DataFrame): variable table for which we want to remove intercorrelated variables.
                Size: N X M  (observations, features).
            rank (pd.Series): Vector of correlation values per feature (of size 1 X M).
            corr_type (str): String specifying the correlation type that we are investigating. 
                Must be 'Pearson' or 'Spearman'.
            thresh_inter_corr (float): Numerical value specifying the threshold above which two variables are 
                considered to be correlated.
            min_n_feat_total (int): Minimum number of features to keep in the table.
        
        Returns:
            pd.DataFrame: Final variable table with the least correlated variables that are kept.
        """
        # Initialization
        n_features = variable_table.shape[1]

        # Compute correlation matrix
        if corr_type == 'Spearman':
            corr_mat = abs(np.corrcoef(variable_table.rank(), rowvar=False))
        elif corr_type == 'Pearson':
            corr_mat = abs(np.corrcoef(variable_table, rowvar=False))
        else:
            raise ValueError('corr_type must be either "Pearson" or "Spearman"')

        # Set diagonal elements to Nans
        np.fill_diagonal(corr_mat, val=np.nan)

        # Calculate mean inter-variable correlation
        mean_corr = np.nanmean(corr_mat, axis=1)
        
        # Looping over all features once
        # rank variables once, for meaningful variable loop.
        ind_loop = pd.Series(mean_corr).rank(method="first") - 1
        # Create a copy of the correlation matrix (to be modified)
        corr_mat_temp = corr_mat.copy()
        while True:
            for f in range(n_features):
                # Use index loop if not NaN
                try:
                    i = int(ind_loop[f])
                except:
                    i = 0
                # Select the row of the current feature
                row = corr_mat_temp[i][:]
                correlated = 1*(row > thresh_inter_corr)  # to turn into integers

                # While the correlations are above the threshold for the select row, we select another row
                while sum(correlated) > 0 and np.isnan(row).sum != len(row):
                    # Find the variable with the highest correlation and drop the one with the lowest rank
                    ind_max = np.nanargmax(row)
                    ind_min = np.nanargmin(np.array([rank[i], rank[ind_max]]))
                    if ind_min == 0:
                        # Drop the current row if the current feature has the lowest correlation with outcome
                        corr_mat_temp[i][:] = np.nan
                        corr_mat_temp[:][i] = np.nan
                        row[:] = np.nan
                    else:
                        # Drop the feature with the highest correlation to the current feature with the lowest correlation with outcome
                        corr_mat_temp[ind_max][:] = np.nan
                        corr_mat_temp[:][ind_max] = np.nan
                        row[ind_max] = np.nan

                    # Update the correlated vector
                    correlated = row > thresh_inter_corr

                # If all the rows are NaN, we keep the variable with the highest rank
                if (1*np.isnan(corr_mat_temp)).sum() == corr_mat_temp.size:
                    ind_keep = np.nanargmax(rank)
                else:
                    ind_keep = list()
                    for row in range(corr_mat_temp.shape[0]):
                        if 1*np.isnan(corr_mat_temp[row][:]).sum() < corr_mat_temp.shape[1]:
                            ind_keep.append(row)

            # if ind_keep happens to be a numpy type convert it to list for better subscripting
            if isinstance(ind_keep, np.int64):
                ind_keep = [ind_keep.tolist()]  # work around
            elif isinstance(ind_keep, np.ndarray):
                ind_keep = ind_keep.tolist()

            # Update threshold if the number of variables is too small or too large
            if len(ind_keep) < min_n_feat_total:
                # Increase the threshold (less stringent)
                thresh_inter_corr = thresh_inter_corr + 0.05
                corr_mat_temp = corr_mat.copy() # reset the correlation matrix
            else:
                break
        
        # Make sure we have the best 
        if len(ind_keep) != min_n_feat_total:
            # Take the features with the highest rank
            ind_keep = sorted(ind_keep)[:min_n_feat_total]

        #  Creating new variable_table
        columns = [variable_table.columns[idx] for idx in ind_keep]
        variable_table = variable_table.loc[:, columns]
        
        return variable_table

    def apply_fda_one_space(
            self, 
            ml: Dict, 
            variable_table: List, 
            outcome_table_binary: pd.DataFrame,
            del_variants: bool = True,
            logging_dict: Dict = None
        ) -> List:
        """
        Applies false discovery avoidance method.

        Args:
            ml (dict): Machine learning dictionary containing the learning options.
            variable_table (List): Table of variables.
            outcome_table_binary (pd.DataFrame): Table of binary outcomes.
            del_variants (bool, optional): If True, will delete the variants of the same feature. Defaults to True.

        Returns:
            List: Table of variables after feature set reduction.
        """
        # Initilization
        n_splits = ml['fSetReduction']['FDA']['nSplits']
        corr_type = ml['fSetReduction']['FDA']['corrType']
        thresh_stable_start = ml['fSetReduction']['FDA']['threshStableStart']
        thresh_inter_corr = ml['fSetReduction']['FDA']['threshInterCorr']
        min_n_feat_stable = ml['fSetReduction']['FDA']['minNfeatStable']
        min_n_feat_total = ml['fSetReduction']['FDA']['minNfeat']
        seed = ml['fSetReduction']['FDA']['seed']

        # Initialization - logging
        if logging_dict is not None:
            table_level = variable_table.Properties['Description'].split('__')[-1]
            logging_dict['one_space']['unstable'][table_level] = {}
            logging_dict['one_space']['inter_corr'][table_level] = {}

        # Getting the correlation table for the radiomics table
        radiomics_table_temp = variable_table.copy()
        outcome_table_binary_temp = outcome_table_binary.copy()
        
        # Get the correlation table
        corr_table = self.__get_fda_corr_table(
            radiomics_table_temp, 
            outcome_table_binary_temp, 
            n_splits, 
            corr_type, 
            seed
        )

        # Calculating the total numbers of features
        feature_total = radiomics_table_temp.shape[1]

        # Cut unstable features (Rmin cut)
        if feature_total > min_n_feat_stable:
            # starting threshold (set by user)
            thresh_stable = thresh_stable_start
            while True:
                # find which features are stable
                var_names_stable, corrs_stable = self.__find_fda_stable(corr_table, thresh_stable)

                # Keep the best textural parameters per image space (deleting variants)
                if del_variants:
                    var_names_stable, corrs_stable = self.__keep_best_text_param(corr_table, var_names_stable, corrs_stable)

                # count the number of stable features
                n_stable = var_names_stable.size
                
                # stop if the minimum number of stable features is reached, if not, lower the threshold.
                if n_stable >= min_n_feat_stable:
                    break
                else:
                    thresh_stable = thresh_stable - 0.05
                
                # stop if the threshold is zero or below
                if thresh_stable <= 0:
                    break
                
            # take the best mean correlation
            if n_stable > min_n_feat_stable:
                var_names_stable, corr_mean_stable = self.__find_fda_best_mean(corrs_stable, min_n_feat_stable)
            else:
                # Compute mean correlation
                corr_mean_stable = corr_table.mean()

            # Finalize radiomics tables before inter-correlation cut
            if len(var_names_stable) > 0:
                var_names = var_names_stable
                if isinstance(radiomics_table_temp, pd.Series):
                    radiomics_table_temp = radiomics_table_temp[[var_names]]
                else:
                    radiomics_table_temp = radiomics_table_temp[var_names]
                radiomics_table_temp = finalize_rad_table(radiomics_table_temp)
            else:
                radiomics_table_temp = pd.DataFrame()
        else:
            # if there is less features than the minimal number, take them all
            n_stable = feature_total

            # Compute mean correlation
            corr_mean_stable = corr_table.mean()

        # Update logging
        if logging_dict is not None:
            logging_dict['one_space']['unstable'][table_level] = radiomics_table_temp.columns.shape[0]

        # Inter-Correlation Cut
        if radiomics_table_temp.shape[1] > 1 and n_stable > min_n_feat_total:
            radiomics_table_temp = self.__remove_correlated_variables(
                radiomics_table_temp, 
                corr_mean_stable.abs(), 
                corr_type, 
                thresh_inter_corr,
                min_n_feat_total
            )
            
            # Finalize radiomics table
            radiomics_table_temp = finalize_rad_table(radiomics_table_temp)
        
        # Update logging
        if logging_dict is not None:
            logging_dict['one_space']['inter_corr'][table_level] = get_full_rad_names(
                radiomics_table_temp.Properties['userData']['variables']['var_def'], 
                radiomics_table_temp.columns.values
            ).tolist()

        return radiomics_table_temp
    
    def apply_fda(
            self, 
            ml: Dict, 
            variable_table: List, 
            outcome_table_binary: pd.DataFrame,
            logging: bool = True,
            path_save_logging: Path = None
        ) -> List:
        """
        Applies false discovery avoidance method.

        Args:
            ml (dict): Machine learning dictionary containing the learning options.
            variable_table (List): Table of variables.
            outcome_table_binary (pd.DataFrame): Table of binary outcomes.
            logging (bool, optional): If True, will save a dict that tracks features selsected for each level. Defaults to True.
            path_save_logging (Path, optional): Path to save the logging dict. Defaults to None.

        Returns:
            List: Table of variables after feature set reduction.
        """
        # Initialization
        rad_tables = variable_table.copy()
        n_rad_tables = len(rad_tables)
        variable_tables = []
        logging_dict = {'one_space': {'unstable': {}, 'inter_corr': {}}, 'final': {}}

        # Apply FDA for each image space/radiomics table
        for r in range(n_rad_tables):
            if logging:
                variable_tables.append(self.apply_fda_one_space(ml, rad_tables[r], outcome_table_binary, logging_dict=logging_dict))
            else:
                variable_tables.append(self.apply_fda_one_space(ml, rad_tables[r], outcome_table_binary))
        
        # Combine radiomics tables
        variable_table = combine_rad_tables(variable_tables)

        # Apply FDA again on the combined radiomics table
        variable_table = self.apply_fda_one_space(ml, variable_table, outcome_table_binary, del_variants=False)

        # Update logging dict
        if logging:
            logging_dict['final'] = get_full_rad_names(variable_table.Properties['userData']['variables']['var_def'], 
                                                       variable_table.columns.values).tolist()
            if path_save_logging is not None:
                path_save_logging = Path(path_save_logging).parent / 'fda_logging_dict.json'
                save_json(path_save_logging, logging_dict, cls=NumpyEncoder)
        
        return variable_table
    
    def apply_fda_balanced(
            self, 
            ml: Dict, 
            variable_table: List, 
            outcome_table_binary: pd.DataFrame,
        ) -> List:
        """
        Applies false discovery avoidance method but balances the number of features on each level.

        Args:
            ml (dict): Machine learning dictionary containing the learning options.
            variable_table (List): Table of variables.
            outcome_table_binary (pd.DataFrame): Table of binary outcomes.
            logging (bool, optional): If True, will save a dict that tracks features selsected for each level. Defaults to True.
            path_save_logging (Path, optional): Path to save the logging dict. Defaults to None.

        Returns:
            List: Table of variables after feature set reduction.
        """
        # Initilization
        rad_tables = variable_table.copy()
        n_rad_tables = len(rad_tables)
        variable_tables_all_levels = []
        levels = [[], [], []]

        # Organize the tables by level
        for r in range(n_rad_tables):
            if 'morph' in rad_tables[r].Properties['Description'].lower():
                levels[0].append(rad_tables[r])
            elif 'intensity' in rad_tables[r].Properties['Description'].lower():
                levels[0].append(rad_tables[r])
            elif 'texture' in rad_tables[r].Properties['Description'].lower():
                levels[0].append(rad_tables[r])
            elif 'mean' in rad_tables[r].Properties['Description'].lower() or \
                 'laws' in rad_tables[r].Properties['Description'].lower() or \
                 'log' in rad_tables[r].Properties['Description'].lower() or \
                 'gabor' in rad_tables[r].Properties['Description'].lower() or \
                 'coif' in rad_tables[r].Properties['Description'].lower() or \
                 'wavelet' in rad_tables[r].Properties['Description'].lower():
                levels[1].append(rad_tables[r])
            elif 'glcm' in rad_tables[r].Properties['Description'].lower():
                levels[2].append(rad_tables[r])

        # Apply FDA for each image space/radiomics table for each level
        for level in levels:
            variable_tables = []
            if len(level) == 0:
                continue
            for r in range(len(level)):
                variable_tables.append(self.apply_fda_one_space(ml, level[r], outcome_table_binary))
            
            # Combine radiomics tables
            variable_table = combine_rad_tables(variable_tables)

            # Apply FDA again on the combined radiomics table
            variable_table = self.apply_fda_one_space(ml, variable_table, outcome_table_binary, del_variants=False)

            # Add-up the tables
            variable_tables_all_levels.append(variable_table)
        
        # Combine radiomics tables of all 3 major levels (original, linear filters and textures)
        variable_table_all_levels = combine_rad_tables(variable_tables_all_levels)

        # Apply FDA again on the combined radiomics table
        variable_table_all_levels = self.apply_fda_one_space(ml, variable_table_all_levels, outcome_table_binary, del_variants=False)
        
        return variable_table_all_levels

    def apply_random_fsr_one_space(
            self,
            ml: Dict,
            variable_table: pd.DataFrame,
        ) -> List:
        seed = ml['fSetReduction']['FDA']['seed']
        
        # Setting the seed
        np.random.seed(seed)

        # Random select 10 columns (features)
        random_df = np.random.choice(variable_table.columns.values.tolist(), 10, replace=False)
        random_df = variable_table[random_df]

        return finalize_rad_table(random_df)
    
    def apply_random_fsr(
            self, 
            ml: Dict, 
            variable_table: List, 
        ) -> List:
        """
        Applies random feature set reduction by choosing a random number of features.

        Args:
            ml (dict): Machine learning dictionary containing the learning options.
            variable_table (List): Table of variables.
            outcome_table_binary (pd.DataFrame): Table of binary outcomes.

        Returns:
            List: Table of variables after feature set reduction.
        """
        # Iinitilization
        rad_tables = variable_table.copy()
        n_rad_tables = len(rad_tables)
        variable_tables = []

        # Apply FDA for each image space/radiomics table
        for r in range(n_rad_tables):
            variable_tables.append(self.apply_random_fsr_one_space(ml, rad_tables[r]))
        
        # Combine radiomics tables
        variable_table = combine_rad_tables(variable_tables)

        # Apply FDA again on the combined radiomics table
        variable_table = self.apply_random_fsr_one_space(ml, variable_table)
        
        return variable_table
    
    def apply_fsr(self, ml: Dict, variable_table: List, outcome_table_binary: pd.DataFrame, path_save_logging: Path = None) -> List:
        """
        Applies feature set reduction method.

        Args:
            ml (dict): Machine learning dictionary containing the learning options.
            variable_table (List): Table of variables.
            outcome_table_binary (pd.DataFrame): Table of binary outcomes.

        Returns:
            List: Table of variables after feature set reduction.
        """
        if self.method.lower() == "fda":
            variable_table = self.apply_fda(ml, variable_table, outcome_table_binary, path_save_logging=path_save_logging)
        elif self.method.lower() == "fdabalanced":
            variable_table = self.apply_fda_balanced(ml, variable_table, outcome_table_binary)
        elif self.method.lower() == "random":
            variable_table = self.apply_random_fsr(ml, variable_table)
        elif self.method == "LASSO":
            raise NotImplementedError("LASSO not implemented yet.")
        elif self.method == "mRMR":
            raise NotImplementedError("mRMR not implemented yet.")
        else:
            raise ValueError("FSR method is None or unknown: " + self.method)
        return variable_table



import logging
import math
import os
import pickle
import sys
from copy import deepcopy
from datetime import datetime
from itertools import product
from pathlib import Path
from time import time
from typing import Dict, List, Union

import numpy as np
import pandas as pd
import ray
from tqdm import trange

import MEDiml


class BatchExtractorTexturalFilters(object):
    """
    Organizes all the patients/scans in batches to extract all the radiomic features
    """

    def __init__(
            self, 
            path_read: Union[str, Path],
            path_csv: Union[str, Path],
            path_params: Union[str, Path],
            path_save: Union[str, Path],
            n_batch: int = 4
    ) -> None:
        """
        constructor of the BatchExtractor class 
        """
        self._path_csv = Path(path_csv)
        self._path_params = Path(path_params)
        self._path_read = Path(path_read)
        self._path_save = Path(path_save)
        self.roi_types = []
        self.roi_type_labels = []
        self.n_bacth = n_batch
        self.glcm_features = [
            "Fcm_joint_max",
            "Fcm_joint_avg",
            "Fcm_joint_var",
            "Fcm_joint_entr",
            "Fcm_diff_avg",
            "Fcm_diff_var",
            "Fcm_diff_entr",
            "Fcm_sum_avg",
            "Fcm_sum_var",
            "Fcm_sum_entr",
            "Fcm_energy",
            "Fcm_contrast",
            "Fcm_dissimilarity",
            "Fcm_inv_diff",
            "Fcm_inv_diff_norm",
            "Fcm_inv_diff_mom",
            "Fcm_inv_diff_mom_norm",
            "Fcm_inv_var",
            "Fcm_corr",
            "Fcm_auto_corr",
            "Fcm_clust_tend",
            "Fcm_clust_shade",
            "Fcm_clust_prom",
            "Fcm_info_corr1",
            "Fcm_info_corr2"
        ]

    def __load_and_process_params(self) -> Dict:
        """Load and process the computing & batch parameters from JSON file"""
        # Load json parameters
        im_params = MEDiml.utils.json_utils.load_json(self._path_params)
        
        # Update class attributes
        self.roi_types.extend(im_params['roi_types'])
        self.roi_type_labels.extend(im_params['roi_type_labels'])
        self.n_bacth = im_params['n_batch'] if 'n_batch' in im_params else self.n_bacth

        return im_params

    def __compute_radiomics_one_patient(
            self,
            name_patient: str,
            roi_name: str,
            im_params: Dict,
            roi_type: str,
            roi_type_label: str,
            log_file: Union[Path, str],
            skip_existing: bool
        ) -> str:
        """
        Computes all radiomics features (Texture & Non-texture) for one patient/scan

        Args:
            name_patient(str): scan or patient full name. It has to respect the MEDiml naming convention:
                PatientID__ImagingScanName.ImagingModality.npy
            roi_name(str): name of the ROI that will  be used in computation.
            im_params(Dict): Dict of parameters/settings that will be used in the processing and computation.
            roi_type(str): Type of ROI used in the processing and computation (for identification purposes)
            roi_type_label(str): Label of the ROI used, to make it identifiable from other ROIs.
            log_file(Union[Path, str]): Path to the logging file.
            skip_existing(bool): True to skip the computation of the features for the scans that already have been computed.

        Returns:
            Union[Path, str]: Path to the updated logging file.
        """
        # Check if the features for the current filter have already been computed
        if skip_existing:
            list_feature = []
            # Find the glcm filters that have not been computed yet
            for i in range(len(self.glcm_features)):
                index_dot = name_patient.find('.')
                ext = name_patient.find('.npy')
                name_save = name_patient[:index_dot] + '(' + roi_type_label + ')' + name_patient[index_dot : ext] + ".json"
                name_roi_type = roi_type + '_' + self.glcm_features[i]
                path_to_check = Path(self._path_save / f'features({name_roi_type})')
                if not (path_to_check / name_save).exists():
                    list_feature.append(i)
            # If all the features have already been computed, skip the computation
            if len(list_feature) == 0:
                return log_file
        
        # Setting up logging settings
        logging.basicConfig(filename=log_file, level=logging.DEBUG, force=True)

        # start timer
        t_start = time()

        # Initialization
        message = f"\n***************** COMPUTING FEATURES: {name_patient} *****************"
        logging.info(message)

        # Load MEDscan instance
        try:
            with open(self._path_read / name_patient, 'rb') as f: medscan = pickle.load(f)
            medscan = MEDiml.MEDscan(medscan)
        except Exception as e:
            logging.error(f"\n ERROR LOADING PATIENT {name_patient}:\n {e}")
            return None

        # Init processing & computation parameters
        medscan.init_params(im_params)
        logging.debug('Parameters parsed, json file is valid.')

        # Get ROI (region of interest)
        logging.info("\n--> Extraction of ROI mask:")
        try:
            vol_obj_init, roi_obj_init = MEDiml.processing.get_roi_from_indexes(
                medscan,
                name_roi=roi_name,
                box_string=medscan.params.process.box_string
            )
        except:
            # if for the current scan ROI is not found, computation is aborted. 
            return log_file

        start = time()
        message = '--> Non-texture features pre-processing (interp + re-seg) for "Scale={}"'.\
            format(str(medscan.params.process.scale_non_text[0]))
        logging.info(message)

        # Interpolation
        # Intensity Mask
        vol_obj = MEDiml.processing.interp_volume(
            medscan=medscan,
            vol_obj_s=vol_obj_init,
            vox_dim=medscan.params.process.scale_non_text,
            interp_met=medscan.params.process.vol_interp,
            round_val=medscan.params.process.gl_round,
            image_type='image',
            roi_obj_s=roi_obj_init,
            box_string=medscan.params.process.box_string
        )
        # Morphological Mask
        roi_obj_morph = MEDiml.processing.interp_volume(
            medscan=medscan,
            vol_obj_s=roi_obj_init,
            vox_dim=medscan.params.process.scale_non_text,
            interp_met=medscan.params.process.roi_interp,
            round_val=medscan.params.process.roi_pv,
            image_type='roi', 
            roi_obj_s=roi_obj_init,
            box_string=medscan.params.process.box_string
        )

        # Re-segmentation
        # Intensity mask range re-segmentation
        roi_obj_int = deepcopy(roi_obj_morph)
        roi_obj_int.data = MEDiml.processing.range_re_seg(
            vol=vol_obj.data, 
            roi=roi_obj_int.data,
            im_range=medscan.params.process.im_range
        )
        # Intensity mask outlier re-segmentation
        roi_obj_int.data = np.logical_and(
            MEDiml.processing.outlier_re_seg(
                vol=vol_obj.data, 
                roi=roi_obj_int.data, 
                outliers=medscan.params.process.outliers
            ),
            roi_obj_int.data
        ).astype(int)
        logging.info(f"{time() - start}\n")

        # Reset timer
        start = time()

        # Image textural filtering
        logging.info("--> Image textural filtering:")

        # Preparation of computation :
        medscan.init_ntf_calculation(vol_obj)

        # ROI Extraction :
        try:
            vol_int_re = MEDiml.processing.roi_extract(
                vol=vol_obj.data, 
                roi=roi_obj_int.data
            )
        except Exception as e:
            print(name_patient, e)
            return log_file

        # Apply textural filter
        try:
            if medscan.params.process.user_set_min_value is None:
                medscan.params.process.user_set_min_value = np.nanmin(vol_int_re)
            vol_obj_all_features = MEDiml.filters.apply_filter(
                medscan, 
                vol_int_re, 
                user_set_min_val=medscan.params.process.user_set_min_value
            )
        except Exception as e:
            print(e)
            logging.error(f'PROBLEM WITH TEXTURAL FILTERING: {e}')
            return log_file
        
        # Initialize ray
        if ray.is_initialized():
            ray.shutdown()

        ray.init(local_mode=True, include_dashboard=True, num_cpus=self.n_bacth)

        # Loop through all the filters and extract the features for each filter
        ids = []
        nb_filters = len(list_feature)
        if nb_filters < self.n_bacth:
            self.n_bacth = nb_filters
        for i in range(self.n_bacth):
            # Extract the filtered volume
            filter_idx = list_feature[i]
            vol_obj.data = deepcopy(vol_obj_all_features[...,filter_idx])

            # Compute radiomics features
            logging.info(f"--> Computation of radiomics features for filter {filter_idx}:")

            ids.append(
                self.__compute_radiomics_filtered_volume.remote(
                    self,
                    medscan=medscan,
                    vol_obj=vol_obj,
                    roi_obj_int=roi_obj_int,
                    roi_obj_morph=roi_obj_morph,
                    name_patient=name_patient,
                    roi_name=roi_name,
                    roi_type=roi_type + '_' + self.glcm_features[filter_idx],
                    roi_type_label=roi_type_label,
                    log_file=log_file
                )
            )
        # Distribute the remaining tasks
        nb_job_left = nb_filters - self.n_bacth
        if nb_job_left > 0:
            for i in range(nb_filters - nb_job_left, nb_filters):
                ready, not_ready = ray.wait(ids, num_returns=1)
                ids = not_ready
                try:
                    log_file = ray.get(ready)[0]
                except:
                    pass
                # Extract the filtered volume
                filter_idx = list_feature[i]
                vol_obj.data = deepcopy(vol_obj_all_features[...,filter_idx])

                # Compute radiomics features
                logging.info(f"--> Computation of radiomics features for filter {filter_idx}:")

                ids.append(
                    self.__compute_radiomics_filtered_volume.remote(
                        self,
                        medscan=medscan,
                        vol_obj=vol_obj,
                        roi_obj_int=roi_obj_int,
                        roi_obj_morph=roi_obj_morph,
                        name_patient=name_patient,
                        roi_name=roi_name,
                        roi_type=roi_type + '_' + self.glcm_features[filter_idx],
                        roi_type_label=roi_type_label,
                        log_file=log_file
                    )
                )
        
        logging.info(f"TOTAL TIME:{time() - t_start} seconds\n\n")

        # Empty memory
        del medscan

    @ray.remote
    def __compute_radiomics_filtered_volume(
            self,
            medscan: MEDiml.MEDscan,
            vol_obj,
            roi_obj_int,
            roi_obj_morph,
            name_patient,
            roi_name,
            roi_type,
            roi_type_label,
            log_file
        ) -> Union[Path, str]:

        # time
        t_start = time()

        # ROI Extraction :
        vol_int_re = deepcopy(vol_obj.data)

        # check if ROI is empty
        if math.isnan(np.nanmax(vol_int_re)) and math.isnan(np.nanmin(vol_int_re)):
            logging.error(f'PROBLEM WITH INTENSITY MASK. ROI {roi_name} IS EMPTY.')
            return log_file
        
        # Computation of non-texture features
        logging.info("--> Computation of non-texture features:")

        # Morphological features extraction
        try:
            morph = MEDiml.biomarkers.morph.extract_all(
                vol=vol_obj.data, 
                mask_int=roi_obj_int.data, 
                mask_morph=roi_obj_morph.data,
                res=medscan.params.process.scale_non_text,
                intensity_type=medscan.params.process.intensity_type
            )
        except Exception as e:
            logging.error(f'PROBLEM WITH COMPUTATION OF MORPHOLOGICAL FEATURES {e}')
            morph = None

        # Local intensity features extraction
        try:
            local_intensity = MEDiml.biomarkers.local_intensity.extract_all(
                img_obj=vol_obj.data,
                roi_obj=roi_obj_int.data,
                res=medscan.params.process.scale_non_text,
                intensity_type=medscan.params.process.intensity_type
            )
        except Exception as e:
            logging.error(f'PROBLEM WITH COMPUTATION OF LOCAL INTENSITY FEATURES {e}')
            local_intensity = None

        # statistical features extraction
        try:
            stats = MEDiml.biomarkers.stats.extract_all(
                vol=vol_int_re,
                intensity_type=medscan.params.process.intensity_type
            )
        except Exception as e:
            logging.error(f'PROBLEM WITH COMPUTATION OF STATISTICAL FEATURES {e}')
            stats = None

        # Intensity histogram equalization of the imaging volume
        vol_quant_re, _ = MEDiml.processing.discretize(
            vol_re=vol_int_re,
            discr_type=medscan.params.process.ih['type'], 
            n_q=medscan.params.process.ih['val'], 
            user_set_min_val=medscan.params.process.user_set_min_value
        )
        
        # Intensity histogram features extraction
        try:
            int_hist = MEDiml.biomarkers.intensity_histogram.extract_all(
                vol=vol_quant_re
            )
        except Exception as e:
            logging.error(f'PROBLEM WITH COMPUTATION OF INTENSITY HISTOGRAM FEATURES {e}')
            int_hist = None
        
        # Intensity histogram equalization of the imaging volume
        if medscan.params.process.ivh and 'type' in medscan.params.process.ivh and 'val' in medscan.params.process.ivh:
            if medscan.params.process.ivh['type'] and medscan.params.process.ivh['val']:
                vol_quant_re, wd = MEDiml.processing.discretize(
                    vol_re=vol_int_re,
                    discr_type=medscan.params.process.ivh['type'], 
                    n_q=medscan.params.process.ivh['val'], 
                    user_set_min_val=medscan.params.process.user_set_min_value,
                    ivh=True
                )
        else:
            vol_quant_re = vol_int_re
            wd = 1

        # Intensity volume histogram features extraction
        try:
            int_vol_hist = MEDiml.biomarkers.int_vol_hist.extract_all(
                medscan=medscan,
                vol=vol_quant_re,
                vol_int_re=vol_int_re, 
                wd=wd
            )
        except:
            print("Error ivh:",name_patient)
            int_vol_hist = {'Fivh_V10': [],
                    'Fivh_V90': [],
                    'Fivh_I10': [],
                    'Fivh_I90': [],
                    'Fivh_V10minusV90': [],
                    'Fivh_I10minusI90': [],
                    'Fivh_auc': []
                    }

        # End of Non-Texture features extraction
        logging.info(f"End of non-texture features extraction: {time() - t_start}\n")

        # Computation of texture features
        logging.info("--> Computation of texture features:")

        # Compute radiomics features for each scale text
        count = 0
        logging.info(f"{time() - t_start}\n")
        
        # Compute features for each discretisation algorithm and for each grey-level  
        for a, n in product(range(medscan.params.process.n_algo), range(medscan.params.process.n_gl)):
            count += 1 
            start = time()
            message = '--> Computation of texture features in image ' \
                    'space for "Scale= {}", "Algo={}", "GL={}" ({}):'.format(
                        str(medscan.params.process.scale_text[0][1]),
                        medscan.params.process.algo[a],
                        str(medscan.params.process.gray_levels[a][n]),
                        str(count) + '/' + str(medscan.params.process.n_exp)
                        )
            logging.info(message)

            # Preparation of computation :
            medscan.init_tf_calculation(algo=a, gl=n, scale=0)

            # Discretisation :
            try:
                vol_quant_re, _ = MEDiml.processing.discretize(
                    vol_re=vol_int_re,
                    discr_type=medscan.params.process.algo[a], 
                    n_q=medscan.params.process.gray_levels[a][n], 
                    user_set_min_val=medscan.params.process.user_set_min_value
                )
            except Exception as e:
                logging.error(f'PROBLEM WITH DISCRETIZATION: {e}')
                vol_quant_re = None

            # GLCM features extraction
            try:
                glcm = MEDiml.biomarkers.glcm.extract_all(
                    vol=vol_quant_re, 
                    dist_correction=medscan.params.radiomics.glcm.dist_correction
                )
            except Exception as e:
                logging.error(f'PROBLEM WITH COMPUTATION OF GLCM FEATURES {e}')
                glcm = None

            # GLRLM features extraction
            try:
                glrlm = MEDiml.biomarkers.glrlm.extract_all(
                    vol=vol_quant_re,
                    dist_correction=medscan.params.radiomics.glrlm.dist_correction
                )
            except Exception as e:
                logging.error(f'PROBLEM WITH COMPUTATION OF GLRLM FEATURES {e}')
                glrlm = None

            # GLSZM features extraction
            try:
                glszm = MEDiml.biomarkers.glszm.extract_all(vol=vol_quant_re)
            except Exception as e:
                logging.error(f'PROBLEM WITH COMPUTATION OF GLSZM FEATURES {e}')
                glszm = None

            # GLDZM features extraction
            try:
                gldzm = MEDiml.biomarkers.gldzm.extract_all(
                    vol_int=vol_quant_re, 
                    mask_morph=roi_obj_morph.data
                )
            except Exception as e:
                logging.error(f'PROBLEM WITH COMPUTATION OF GLDZM FEATURES {e}')
                gldzm = None

            # NGTDM features extraction
            try:
                ngtdm = MEDiml.biomarkers.ngtdm.extract_all(
                    vol=vol_quant_re, 
                    dist_correction=medscan.params.radiomics.ngtdm.dist_correction
                )
            except Exception as e:
                logging.error(f'PROBLEM WITH COMPUTATION OF NGTDM FEATURES {e}')
                ngtdm = None

            # NGLDM features extraction
            try:
                ngldm = MEDiml.biomarkers.ngldm.extract_all(vol=vol_quant_re)
            except Exception as e:
                logging.error(f'PROBLEM WITH COMPUTATION OF NGLDM FEATURES {e}')
                ngldm = None
            
            # Update radiomics results class
            medscan.update_radiomics(
                int_vol_hist_features=int_vol_hist, 
                morph_features=morph,
                loc_int_features=local_intensity, 
                stats_features=stats, 
                int_hist_features=int_hist,
                glcm_features=glcm, 
                glrlm_features=glrlm, 
                glszm_features=glszm, 
                gldzm_features=gldzm, 
                ngtdm_features=ngtdm, 
                ngldm_features=ngldm
            )
                
        # End of texture features extraction
        logging.info(f"End of texture features extraction: {time() - start}\n")

        # Saving radiomics results
        medscan.save_radiomics(
            scan_file_name=name_patient,
            path_save=self._path_save,
            roi_type=roi_type,
            roi_type_label=roi_type_label,
        )
        
        logging.info(f"TOTAL TIME 1 FILTER:{time() - t_start} seconds\n\n")

        return log_file
    
    @ray.remote
    def __compute_radiomics_tables(
            self,
            table_tags: List,
            log_file: Union[str, Path],
            im_params: Dict,
            feature_name: str
        ) -> None:
        """
        Creates radiomic tables off of the saved dicts with the computed features and save it as CSV files

        Args:
            table_tags(List): Lists of information about scans, roi type and imaging space (or filter space)
            log_file(Union[str, Path]): Path to logging file.
            im_params(Dict): Dictionary of parameters.
        
        Returns:
            None.
        """
        n_tables = len(table_tags)

        for t in range(0, n_tables):
            scan = table_tags[t][0]
            roi_type = table_tags[t][1]
            roi_label = table_tags[t][2]
            im_space = table_tags[t][3]
            modality = table_tags[t][4]

            # extract parameters for the current modality
            if modality == 'CTscan' and 'imParamCT' in im_params:
                im_params_mod = im_params['imParamCT']
            elif modality== 'MRscan' and 'imParamMR' in im_params:
                im_params_mod = im_params['imParamMR']
            elif modality == 'PTscan' and 'imParamPET' in im_params:
                im_params_mod = im_params['imParamPET']
            # extract name save of the used filter
            if 'filter_type' in im_params_mod:
                filter_type = im_params_mod['filter_type']
                if filter_type in im_params['imParamFilter'] and 'name_save' in im_params['imParamFilter'][filter_type]:
                    name_save = im_params['imParamFilter'][filter_type]['name_save'] + '_' + feature_name
                else:
                    name_save= feature_name
            else:
                name_save= feature_name
            
            # set up table name
            if name_save:
                name_table = 'radiomics__' + scan + \
                '(' + roi_type + ')__'  + name_save + '.npy'   
            else:
                name_table = 'radiomics__' + scan + \
                '(' + roi_type + ')__' + im_space + '.npy'

            # Start timer
            start = time()
            logging.info("\n --> Computing radiomics table: {name_table}...")

            # Wildcard used to look only in the parent folder (save path),
            # no need to recursively look into sub-folders using '**/'.
            wildcard = '*_' + scan + '(' + roi_type + ')*.json'

            # Create radiomics table
            radiomics_table_dict = MEDiml.utils.create_radiomics_table(
                MEDiml.utils.get_file_paths(self._path_save / f'features({roi_label})', wildcard),
                im_space, 
                log_file
            )
            radiomics_table_dict['Properties']['Description'] = name_table

            # Save radiomics table
            save_path = self._path_save / f'features({roi_label})' / name_table
            np.save(save_path, [radiomics_table_dict])

            # Create CSV table and Definitions
            MEDiml.utils.write_radiomics_csv(save_path)

            logging.info(f"DONE\n {time() - start}\n")

        return log_file
    
    def __batch_all_patients(self, im_params: Dict, skip_existing) -> None:
        """
        Create batches of scans to process and compute radiomics features for every single scan.

        Args: 
            im_params(Dict): Dict of the processing & computation parameters.
            skip_existing(bool) : True to skip the computation of the features for the scans that already have been computed.

        Returns:
            None
        """
        # create a batch for each roi type
        n_roi_types = len(self.roi_type_labels)
        for r in range(0, n_roi_types):
            roi_type = self.roi_types[r]
            roi_type_label = self.roi_type_labels[r]
            print(f'\n --> Computing features for the "{roi_type_label}" roi type ...', end = '')

            # READING CSV EXPERIMENT TABLE
            tabel_roi = pd.read_csv(self._path_csv / ('roiNames_' + roi_type_label + '.csv'))
            tabel_roi['under'] = '_'
            tabel_roi['dot'] = '.'
            tabel_roi['npy'] = '.npy'
            name_patients = (pd.Series(
                tabel_roi[['PatientID', 'under', 'under',
                        'ImagingScanName',
                        'dot',
                        'ImagingModality',
                        'npy']].fillna('').values.tolist()).str.join('')).tolist()
            tabel_roi = tabel_roi.drop(columns=['under', 'under', 'dot', 'npy'])
            roi_names = tabel_roi.ROIname.tolist()

            # INITIALIZATION
            os.chdir(self._path_save)
            name_bacth_log = 'batchLog_' + roi_type_label
            p = Path.cwd().glob('*')
            files = [x for x in p if x.is_dir()]
            n_files = len(files)
            exist_file = name_bacth_log in [x.name for x in files]
            if exist_file and (n_files > 0):
                for i in range(0, n_files):
                    if (files[i].name == name_bacth_log):
                        mod_timestamp = datetime.fromtimestamp(
                            Path(files[i]).stat().st_mtime)
                        date = mod_timestamp.strftime("%d-%b-%Y_%HH%MM%SS")
                        new_name = name_bacth_log+'_'+date
                        if sys.platform == 'win32':
                            os.system('move ' + name_bacth_log + ' ' + new_name)
                        else:
                            os.system('mv ' + name_bacth_log + ' ' + new_name)

            os.makedirs(name_bacth_log, 0o777, True)
            path_batch = Path.cwd() / name_bacth_log

            # PRODUCE BATCH COMPUTATIONS
            n_patients = len(name_patients)

            # Produce a list log_file path.
            log_files = [path_batch / ('log_file_' + str(i) + '.log') for i in range(n_patients)]

            # Features computation for each patient (patients loop)
            for i in trange(n_patients):
                self.__compute_radiomics_one_patient(
                    name_patients[i],
                    roi_names[i], 
                    im_params,
                    roi_type,
                    roi_type_label,
                    log_files[i],
                    skip_existing
                )

            print('DONE')

    def __batch_all_tables(self, im_params: Dict):
        """
        Create batches of tables of the extracted features for every imaging scan type (CT, PET...).

        Args: 
            im_params(Dict): Dictionary of parameters.

        Returns:
            None
        """
        # INITIALIZATION
        os.chdir(self._path_save)
        name_batch_log = 'batchLog_tables'
        p = Path.cwd().glob('*')
        files = [x for x in p if x.is_dir()]
        n_files = len(files)
        exist_file = name_batch_log in [x.name for x in files]
        if exist_file and (n_files > 0):
            for i in range(0, n_files):
                if files[i].name == name_batch_log:
                    mod_timestamp = datetime.fromtimestamp(
                        Path(files[i]).stat().st_mtime)
                    date = mod_timestamp.strftime("%d-%b-%Y_%H:%M:%S")
                    new_name = name_batch_log+'_'+date
                    if sys.platform == 'win32':
                        os.system('move ' + name_batch_log + ' ' + new_name)
                    else:
                        os.system('mv ' + name_batch_log + ' ' + new_name)

        os.makedirs(name_batch_log, 0o777, True)
        path_batch = Path.cwd()
        
        # GETTING COMBINATIONS OF scan, roi_type and imageSpaces
        n_roi_types = len(self.roi_type_labels)
        
        # Get all scan names present for the given roi_type_label
        for f_idx in range(0, len(self.glcm_features)):
            # RE-INITIALIZATION
            table_tags = []
            for r in range(0, n_roi_types):
                label = self.roi_type_labels[r]
                wildcard = '*' + label + '*.json'
                roi_type = self.roi_types[r] + '_' + self.glcm_features[f_idx]
                file_paths = MEDiml.utils.get_file_paths(self._path_save / f'features({roi_type})', wildcard)
                n_files = len(file_paths)
                scans = [0] * n_files
                modalities = [0] * n_files
                for f in range(0, n_files):
                    rad_file_name = file_paths[f].stem
                    scans[f] = MEDiml.utils.get_scan_name_from_rad_name(rad_file_name)
                    modalities[f] = rad_file_name.split('.')[1]
                scans = s = (np.unique(np.array(scans))).tolist()
                n_scans = len(scans)
                # Get all scan names present for the given roi_type_label and scans
                for s in range(0, n_scans):
                    scan = scans[s]
                    modality = modalities[s]
                    wildcard = '*' + scan + '(' + label + ')*.json'
                    file_paths = MEDiml.utils.get_file_paths(self._path_save / f'features({roi_type})', wildcard)
                    n_files = len(file_paths)

                    # Finding the images spaces for a test file (assuming that all
                    # files for a given scan and roi_type_label have the same image spaces
                    radiomics = MEDiml.utils.json_utils.load_json(file_paths[0])
                    im_spaces = [key for key in radiomics.keys()]
                    im_spaces = im_spaces[:-1]
                    n_im_spaces = len(im_spaces)
                    # Constructing the table_tags variable
                    for i in range(0, n_im_spaces):
                        im_space = im_spaces[i]
                        table_tags = table_tags + [[scan, label, roi_type, im_space, modality]]

            # PRODUCE BATCH COMPUTATIONS
            n_tables = len(table_tags)
            self.n_bacth = self.n_bacth
            if self.n_bacth is None or self.n_bacth < 0:
                self.n_bacth = 1
            elif n_tables < self.n_bacth:
                self.n_bacth = n_tables

            # Produce a list log_file path.
            log_files = [path_batch / ('log_file_' + str(i) + '.txt') for i in range(self.n_bacth)]

            # Initialize ray
            if ray.is_initialized():
                ray.shutdown()

            ray.init(local_mode=True, include_dashboard=True, num_cpus=self.n_bacth)

            # Distribute the first tasks to all workers
            ids = [self.__compute_radiomics_tables.remote(
                                    self, 
                                    [table_tags[i]], 
                                    log_files[i],
                                    im_params,
                                    self.glcm_features[f_idx])
                    for i in range(self.n_bacth)]

            nb_job_left = n_tables - self.n_bacth

            for _ in trange(n_tables):
                ready, not_ready = ray.wait(ids, num_returns=1)
                ids = not_ready

                # We verify if error has occur during the process
                log_file = ray.get(ready)[0]

                # Distribute the remaining tasks
                if nb_job_left > 0:
                    idx = n_tables - nb_job_left
                    ids.extend([self.__compute_radiomics_tables.remote(
                                    self,
                                    [table_tags[idx]], 
                                    log_file,
                                    im_params,
                                    self.glcm_features[f_idx])])
                    nb_job_left -= 1

        print('DONE')

    def compute_radiomics(self, create_tables: bool = True, skip_existing: bool = False) -> None:
        """Compute all radiomic features for all scans in the CSV file (set in initialization) and organize it
        in JSON and CSV files

        Args:
            create_tables(bool) : True to create CSV tables for the extracted features and not save it in JSON only.
            skip_existing(bool) : True to skip the computation of the features for the scans that already have been computed.
        
        Returns:
            None.
        """

        # Load and process computing parameters
        im_params = self.__load_and_process_params()

        # Batch all scans from CSV file and compute radiomics for each scan
        self.__batch_all_patients(im_params, skip_existing)

        # Create a CSV file off of the computed features for all the scans
        if create_tables:
            self.__batch_all_tables(im_params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List

import numpy as np


def get_patient_names(roi_names: np.ndarray) -> List[str]:
    """Generates all file names for scans using CSV data. 

    Args:
        roi_names (ndarray): Array with CSV data organized as follows 
            [[patient_id], [imaging_scan_name], [imagning_modality]]
        
    Returns:
        list[str]: List of scans files name.
    """
    n_names = np.size(roi_names[0])
    patient_names = [0] * n_names
    for n in range(0, n_names):
        patient_names[n] = roi_names[0][n]+'__'+roi_names[1][n] + \
            '.'+roi_names[2][n]+'.npy'

    return patient_names



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from scipy.stats import iqr, kurtosis, skew, scoreatpercentile, variation


def extract_all(vol: np.ndarray, intensity_type: str) -> dict:
    """Computes Intensity-based statistical features.
    These features refer to "Intensity-based statistical features" (ID = UHIW) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution).
        intensity_type (str): Type of intensity to compute. Can be "arbitrary", "definite" or "filtered".
            Will compute features only for "definite" intensity type.

    Return:
        dict: Dictionnary containing all stats features.

    Raises:
        ValueError: If `intensity_type` is not "arbitrary", "definite" or "filtered".
    """
    assert intensity_type in ["arbitrary", "definite", "filtered"], \
        "intensity_type must be 'arbitrary', 'definite' or 'filtered'"
    
    x = vol[~np.isnan(vol[:])]  # Initialization

    # Initialization of final structure (Dictionary) containing all features.
    stats = {'Fstat_mean': [],
             'Fstat_var': [],
             'Fstat_skew': [],
             'Fstat_kurt': [],
             'Fstat_median': [],
             'Fstat_min': [],
             'Fstat_P10': [],
             'Fstat_P90': [],
             'Fstat_max': [],
             'Fstat_iqr': [],
             'Fstat_range': [],
             'Fstat_mad': [],
             'Fstat_rmad': [],
             'Fstat_medad': [],
             'Fstat_cov': [],
             'Fstat_qcod': [],
             'Fstat_energy': [],
             'Fstat_rms': []
             }

    # STARTING COMPUTATION
    if intensity_type == "definite":
        stats['Fstat_mean'] = np.mean(x)  # Mean
        stats['Fstat_var'] = np.var(x)  # Variance
        stats['Fstat_skew'] = skew(x)  # Skewness
        stats['Fstat_kurt'] = kurtosis(x)  # Kurtosis
        stats['Fstat_median'] = np.median(x)  # Median
        stats['Fstat_min'] = np.min(x)  # Minimum grey level
        stats['Fstat_P10'] = scoreatpercentile(x, 10)  # 10th percentile
        stats['Fstat_P90'] = scoreatpercentile(x, 90)  # 90th percentile
        stats['Fstat_max'] = np.max(x)  # Maximum grey level
        stats['Fstat_iqr'] = iqr(x)  # Interquantile range
        stats['Fstat_range'] = np.ptp(x)  # Range max(x) - min(x)
        stats['Fstat_mad'] = np.mean(np.absolute(x - np.mean(x)))  # Mean absolute deviation
        x_10_90 = x[np.where((x >= stats['Fstat_P10']) &
                                (x <= stats['Fstat_P90']), True, False)]
        stats['Fstat_rmad'] = np.mean(np.abs(x_10_90 - np.mean(x_10_90)))  # Robust mean absolute deviation
        stats['Fstat_medad'] = np.mean(np.absolute(x - np.median(x)))  # Median absolute deviation
        stats['Fstat_cov'] = variation(x)  # Coefficient of variation
        x_75_25 = scoreatpercentile(x, 75) + scoreatpercentile(x, 25)
        stats['Fstat_qcod'] = iqr(x)/x_75_25  # Quartile coefficient of dispersion
        stats['Fstat_energy'] = np.sum(np.power(x, 2))  # Energy
        stats['Fstat_rms'] = np.sqrt(np.mean(np.power(x, 2)))  # Root mean square

    return stats

def mean(vol: np.ndarray) -> float:
    """Computes statistical mean feature of the input dataset (3D Array).
    This feature refers to "Fstat_mean" (ID = Q4LE)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: Statistical mean feature
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.mean(x)  # Mean

def var(vol: np.ndarray) -> float:
    """Computes statistical variance feature of the input dataset (3D Array).
    This feature refers to "Fstat_var" (ID = ECT3)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: Statistical variance feature
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.var(x)  # Variance

def skewness(vol: np.ndarray) -> float:
    """Computes the sample skewness feature of the input dataset (3D Array).
    This feature refers to "Fstat_skew" (ID = KE2A)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: The skewness feature of values along an axis. Returning 0 where all values are
        equal.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return skew(x)  # Skewness

def kurt(vol: np.ndarray) -> float:
    """Computes the kurtosis (Fisher or Pearson) feature of the input dataset (3D Array).
    This feature refers to "Fstat_kurt" (ID = IPH6)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: The kurtosis feature of values along an axis. If all values are equal,
        return -3 for Fisher's definition and 0 for Pearson's definition.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return kurtosis(x)  # Kurtosis

def median(vol: np.ndarray) -> float:
    """Computes the median feature along the specified axis of the input dataset (3D Array).
    This feature refers to "Fstat_median" (ID = Y12H)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: The median feature of the array elements.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.median(x)  # Median

def min(vol: np.ndarray) -> float:
    """Computes the minimum grey level feature of the input dataset (3D Array).
    This feature refers to "Fstat_min" (ID = 1GSF)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: The minimum grey level feature of the array elements.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.min(x)  # Minimum grey level

def p10(vol: np.ndarray) -> float:
    """Computes the score at the 10th percentile feature of the input dataset (3D Array).
    This feature refers to "Fstat_P10" (ID = QG58)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: Score at 10th percentil.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return scoreatpercentile(x, 10)  # 10th percentile

def p90(vol: np.ndarray) -> float:
    """Computes the score at the 90th percentile feature of the input dataset (3D Array).
    This feature refers to "Fstat_P90" (ID = 8DWT)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: Score at 90th percentil.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return scoreatpercentile(x, 90)  # 90th percentile

def max(vol: np.ndarray) -> float:
    """Computes the maximum grey level feature of the input dataset (3D Array).
    This feature refers to "Fstat_max" (ID = 84IY)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: The maximum grey level feature of the array elements.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.max(x)  # Maximum grey level

def iqrange(vol: np.ndarray) -> float:
    """Computes the interquartile range feature of the input dataset (3D Array).
    This feature refers to "Fstat_iqr" (ID = SALO)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: Interquartile range. If axis != None, the output data-type is the same as that of the input.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return iqr(x)  # Interquartile range

def range(vol: np.ndarray) -> float:
    """Range of values (maximum - minimum) feature along an axis of the input dataset (3D Array).
    This feature refers to "Fstat_range" (ID = 2OJQ)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: A new array holding the range of values, unless out was specified,
        in which case a reference to out is returned.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.ptp(x)  # Range max(x) - min(x) 

def mad(vol: np.ndarray) -> float:
    """Mean absolute deviation feature of the input dataset (3D Array).
    This feature refers to "Fstat_mad" (ID = 4FUA)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float : A new array holding mean absolute deviation feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.mean(np.absolute(x - np.mean(x)))  # Mean absolute deviation

def rmad(vol: np.ndarray) -> float:
    """Robust mean absolute deviation feature of the input dataset (3D Array).
    This feature refers to "Fstat_rmad" (ID = 1128)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
        P10(ndarray): Score at 10th percentil.
        P90(ndarray): Score at 90th percentil.

    Returns:
        float: A new array holding the robust mean absolute deviation.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization
    P10 = scoreatpercentile(x, 10)  # 10th percentile
    P90 = scoreatpercentile(x, 90)  # 90th percentile
    x_10_90 = x[np.where((x >= P10) &
                         (x <= P90), True, False)]  # Holding x for (x >= P10) and (x<= P90)
                         
    return np.mean(np.abs(x_10_90 - np.mean(x_10_90)))  # Robust mean absolute deviation

def medad(vol: np.ndarray) -> float:
    """Median absolute deviation feature of the input dataset (3D Array).
    This feature refers to "Fstat_medad" (ID = N72L)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: A new array holding the median absolute deviation feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.mean(np.absolute(x - np.median(x)))  # Median absolute deviation

def cov(vol: np.ndarray) -> float:
    """Computes the coefficient of variation feature of the input dataset (3D Array).
    This feature refers to "Fstat_cov" (ID = 7TET)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: A new array holding the coefficient of variation feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return variation(x)  # Coefficient of variation

def qcod(vol: np.ndarray) -> float:
    """Computes the quartile coefficient of dispersion feature of the input dataset (3D Array).
    This feature refers to "Fstat_qcod" (ID = 9S40)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: A new array holding the quartile coefficient of dispersion feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization
    x_75_25 = scoreatpercentile(x, 75) + scoreatpercentile(x, 25)  

    return iqr(x) / x_75_25  # Quartile coefficient of dispersion

def energy(vol: np.ndarray) -> float:
    """Computes the energy feature of the input dataset (3D Array).
    This feature refers to "Fstat_energy" (ID = N8CA)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: A new array holding the energy feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.sum(np.power(x, 2))  # Energy

def rms(vol: np.ndarray) -> float:
    """Computes the root mean square feature of the input dataset (3D Array).
    This feature refers to "Fstat_rms" (ID = 5ZWQ)  in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)

    Returns:
        float: A new array holding the root mean square feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.sqrt(np.mean(np.power(x, 2)))  # Root mean square



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List, Tuple, Union

import numpy as np

from ..utils.strfind import strfind


def parse_contour_string(contour_string) -> Union[Tuple[float, List[str]],
                                                  Tuple[int, List[str]],
                                                  Tuple[List[int], List[str]]]:
    """Finds the delimeters (:math:`'+'` and :math:`'-'`) and the contour indexe(s) from the given string.

    Args:
        contour_string (str, float or int): Index or string of indexes with
        delimeters. For example: :math:`'3'` or :math:`'1-3+2'`.

    Returns:
        float, int: If ``contour_string`` is a an int or float we return ``contour_string``.
        List[str]: List of the delimeters.
        List[int]: List of the contour indexes.

    Example:
        >>> ``contour_string`` = '1-3+2'
        >>> :function: parse_contour_string(contour_string)
        [1, 2, 3], ['+', '-']
        >>> ``contour_string`` = 1
        >>> :function: parse_contour_string(contour_string)
        1, []
    """

    if isinstance(contour_string, (int, float)):
        return contour_string, []

    ind_plus = strfind(string=contour_string, pattern='\+')
    ind_minus = strfind(string=contour_string, pattern='\-')
    ind_operations = np.sort(np.hstack((ind_plus, ind_minus))).astype(int)

    # Parsing operations and contour numbers
    # AZ: I assume that contour_number is an integer
    if ind_operations.size == 0:
        operations = []
        contour_number = [int(contour_string)]
    else:
        n_op = len(ind_operations)
        operations = [contour_string[ind_operations[i]] for i in np.arange(n_op)]

        contour_number = np.zeros(n_op + 1, dtype=int)
        contour_number[0] = int(contour_string[0:ind_operations[0]])
        for c in np.arange(start=1, stop=n_op):
            contour_number[c] = int(contour_string[(ind_operations[c-1]+1) : ind_operations[c]])

        contour_number[-1] = int(contour_string[(ind_operations[-1]+1):])
        contour_number.tolist()

    return contour_number, operations



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Dict, List, Union

import numpy as np
import skimage.measure as skim


def get_matrix(roi_only: np.ndarray, 
                     levels: Union[np.ndarray, List]) -> Dict:
    r"""
    This function computes the Gray-Level Size Zone Matrix (GLSZM) of the
    region of interest (ROI) of an input volume. The input volume is assumed
    to be isotropically resampled. The zones of different sizes are computed
    using 26-voxel connectivity.
    This matrix refers to "Grey level size zone based features" (ID = 9SAK)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_. 

    Note:
        This function is compatible with 2D analysis (language not adapted in the text).

    Args:
        roi_only_int (ndarray): Smallest box containing the ROI, with the imaging data ready
            for texture analysis computations. Voxels outside the ROI are
            set to NaNs.
        levels (ndarray or List): Vector containing the quantized gray-levels
            in the tumor region (or reconstruction ``levels`` of quantization).

    Returns:
        ndarray: Array of Gray-Level Size Zone Matrix of ``roi_only``.

    REFERENCES:
        [1] Thibault, G., Fertil, B., Navarro, C., Pereira, S., Cau, P., Levy,
        N., Mari, J.-L. (2009). Texture Indexes and Gray Level Size Zone
        Matrix. Application to Cell Nuclei Classification. In Pattern
        Recognition and Information Processing (PRIP) (pp. 140–145).
    """

    # PRELIMINARY
    roi_only = roi_only.copy()
    n_max = np.sum(~np.isnan(roi_only))
    level_temp = np.max(levels) + 1
    roi_only[np.isnan(roi_only)] = level_temp
    levels = np.append(levels, level_temp)

    # QUANTIZATION EFFECTS CORRECTION
    # In case (for example) we initially wanted to have 64 levels, but due to
    # quantization, only 60 resulted.
    unique_vect = levels
    n_l = np.size(levels) - 1

    # INITIALIZATION
    # THIS NEEDS TO BE CHANGED. THE ARRAY INITIALIZED COULD BE TOO BIG!
    glszm = np.zeros((n_l, n_max))

    # COMPUTATION OF glszm
    temp = roi_only.copy().astype('int')
    for i in range(1, n_l+1):
        temp[roi_only != unique_vect[i-1]] = 0
        temp[roi_only == unique_vect[i-1]] = 1
        conn_objects, n_zone = skim.label(temp, return_num=True)
        for j in range(1, n_zone+1):
            col = np.sum(conn_objects == j)
            glszm[i-1, col-1] = glszm[i-1, col-1] + 1

    # REMOVE UNECESSARY COLUMNS
    stop = np.nonzero(np.sum(glszm, 0))[0][-1]
    glszm = np.delete(glszm, range(stop+1, np.shape(glszm)[1]), 1)

    return glszm

def extract_all(vol: np.ndarray,
                glszm: np.ndarray = None) -> Dict:
    """Computes glszm features.
    These features refer to "Grey level size zone based features" (ID = 9SAK)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__. 
    
    Args:
        vol (ndarray): 3D volume, isotropically resampled, quantized
            (e.g. n_g = 32, levels = [1, ..., n_g]),
            with NaNs outside the region of interest.
    
    Returns:
        Dict: Dict of glszm features.
        
    """
    glszm_features = {'Fszm_sze': [],
                      'Fszm_lze': [],
                      'Fszm_lgze': [],
                      'Fszm_hgze': [],
                      'Fszm_szlge': [],
                      'Fszm_szhge': [],
                      'Fszm_lzlge': [],
                      'Fszm_lzhge': [],
                      'Fszm_glnu': [],
                      'Fszm_glnu_norm': [],
                      'Fszm_zsnu': [],
                      'Fszm_zsnu_norm': [],
                      'Fszm_z_perc': [],
                      'Fszm_gl_var': [],
                      'Fszm_zs_var': [],
                      'Fszm_zs_entr': []}

    # GET THE GLSZM MATRIX
    # Correct definition, without any assumption
    vol = vol.copy()
    levels = np.arange(1, np.max(vol[~np.isnan(vol[:])])+1)
    if glszm is None:
        glszm = get_matrix(vol, levels)
    n_s = np.sum(glszm)
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    r_vect = range(1, sz[0]+1)  # Column vectors
    # Column and row indicators for each entry of the glszm
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)
    pg = np.transpose(np.sum(glszm, 1))  # Gray-Level Vector
    pz = np.sum(glszm, 0)  # Zone Size Vector

    # COMPUTING TEXTURES

    # Small zone emphasis
    glszm_features['Fszm_sze'] = (np.matmul(pz, np.transpose(np.power(1.0/np.array(c_vect), 2))))

    # Large zone emphasis
    glszm_features['Fszm_lze'] = (np.matmul(pz, np.transpose(np.power(np.array(c_vect), 2))))

    # Low grey level zone emphasis
    glszm_features['Fszm_lgze'] = np.matmul(pg, np.transpose(np.power(
        1.0/np.array(r_vect), 2)))

    # High grey level zone emphasis
    glszm_features['Fszm_hgze'] = np.matmul(pg, np.transpose(np.power(np.array(r_vect), 2)))

    # Small zone low grey level emphasis
    glszm_features['Fszm_szlge'] = np.sum(np.sum(glszm*(np.power(1.0/r_mat, 2))*(np.power(1.0/c_mat, 2))))

    # Small zone high grey level emphasis
    glszm_features['Fszm_szhge'] = np.sum(np.sum(glszm*(np.power(r_mat, 2))*(np.power(1.0/c_mat, 2))))

    # Large zone low grey levels emphasis
    glszm_features['Fszm_lzlge'] = np.sum(np.sum(glszm*(np.power(1.0/r_mat, 2))*(np.power(c_mat, 2))))

    # Large zone high grey level emphasis
    glszm_features['Fszm_lzhge'] = np.sum(np.sum(glszm*(np.power(r_mat, 2))*(np.power(c_mat, 2))))

    # Gray level non-uniformity
    glszm_features['Fszm_glnu'] = np.sum(np.power(pg, 2)) * n_s

    # Gray level non-uniformity normalised
    glszm_features['Fszm_glnu_norm'] = np.sum(np.power(pg, 2))

    # Zone size non-uniformity
    glszm_features['Fszm_zsnu'] = np.sum(np.power(pz, 2)) * n_s

    # Zone size non-uniformity normalised
    glszm_features['Fszm_zsnu_norm'] = np.sum(np.power(pz, 2))

    # Zone percentage
    glszm_features['Fszm_z_perc'] = np.sum(pg)/(np.matmul(pz, np.transpose(c_vect)))

    # Grey level variance
    temp = r_mat * glszm
    u = np.sum(temp)
    temp = (np.power(r_mat - u, 2)) * glszm
    glszm_features['Fszm_gl_var'] = np.sum(temp)

    # Zone size variance
    temp = c_mat * glszm
    u = np.sum(temp)
    temp = (np.power(c_mat - u, 2)) * glszm
    glszm_features['Fszm_zs_var'] = np.sum(temp)

    # Zone size entropy
    val_pos = glszm[np.nonzero(glszm)]
    temp = val_pos * np.log2(val_pos)
    glszm_features['Fszm_zs_entr'] = -np.sum(temp)

    return glszm_features

def get_single_matrix(vol: np.ndarray) -> np.ndarray:
    """Computes gray level size zone matrix in order to compute the single features.

    Args:
        vol_int: 3D volume, isotropically resampled, 
            quantized (e.g. n_g = 32, levels = [1, ..., n_g]), 
            with NaNs outside the region of interest.
        levels: Vector containing the quantized gray-levels 
            in the tumor region (or reconstruction ``levels`` of quantization).
    
    Returns:
        ndarray: Array of Gray-Level Size Zone Matrix of 'vol'.

    """
    # Correct definition, without any assumption
    vol = vol.copy()
    levels = np.arange(1, np.max(vol[~np.isnan(vol[:])])+1)

    # GET THE gldzm MATRIX
    glszm = get_matrix(vol, levels)

    return glszm

def sze(glszm: np.ndarray) -> float:
    """Computes small zone emphasis feature.
    This feature refers to "Fszm_sze" (ID = 5QRC) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the small zone emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    pz = np.sum(glszm, 0)  # Zone Size Vector

    # Small zone emphasis
    return (np.matmul(pz, np.transpose(np.power(1.0/np.array(c_vect), 2))))

def lze(glszm: np.ndarray) -> float:
    """Computes large zone emphasis feature.
    This feature refers to "Fszm_lze" (ID = 48P8) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the large zone emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    pz = np.sum(glszm, 0)  # Zone Size Vector

    # Large zone emphasis
    return (np.matmul(pz, np.transpose(np.power(np.array(c_vect), 2))))

def lgze(glszm: np.ndarray) -> float:
    """Computes low grey zone emphasis feature.
    This feature refers to "Fszm_lgze" (ID = XMSY) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the low grey zone emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    r_vect = range(1, sz[0]+1)  # Column vectors
    pg = np.transpose(np.sum(glszm, 1))  # Gray-Level Vector

    # Low grey zone emphasis
    return np.matmul(pg, np.transpose(np.power(
        1.0/np.array(r_vect), 2)))

def hgze(glszm: np.ndarray) -> float:
    """Computes high grey zone emphasis feature.
    This feature refers to "Fszm_hgze" (ID = 5GN9) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the high grey zone emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    r_vect = range(1, sz[0]+1)  # Column vectors
    pg = np.transpose(np.sum(glszm, 1))  # Gray-Level Vector

    # High grey zone emphasis
    return np.matmul(pg, np.transpose(np.power(np.array(r_vect), 2)))

def szlge(glszm: np.ndarray) -> float:
    """Computes small zone low grey level emphasis feature.
    This feature refers to "Fszm_szlge" (ID = 5RAI) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the small zone low grey level emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    r_vect = range(1, sz[0]+1)  # Column vectors
    # Column and row indicators for each entry of the glszm
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)

    # Small zone low grey level emphasis
    return np.sum(np.sum(glszm*(np.power(1.0/r_mat, 2))*(np.power(1.0/c_mat, 2))))

def szhge(glszm: np.ndarray) -> float:
    """Computes small zone high grey level emphasis feature.
    This feature refers to "Fszm_szhge" (ID = HW1V) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the small zone high grey level emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    r_vect = range(1, sz[0]+1)  # Column vectors
    # Column and row indicators for each entry of the glszm
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)

    # Small zone high grey level emphasis
    return np.sum(np.sum(glszm*(np.power(r_mat, 2))*(np.power(1.0/c_mat, 2))))

def lzlge(glszm: np.ndarray) -> float:
    """Computes large zone low grey level emphasis feature.
    This feature refers to "Fszm_lzlge" (ID = YH51) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the large zone low grey level emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    r_vect = range(1, sz[0]+1)  # Column vectors
    # Column and row indicators for each entry of the glszm
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)

    # Lage zone low grey level emphasis
    return np.sum(np.sum(glszm*(np.power(1.0/r_mat, 2))*(np.power(c_mat, 2))))

def lzhge(glszm: np.ndarray) -> float:
    """Computes large zone high grey level emphasis feature.
    This feature refers to "Fszm_lzhge" (ID = J17V) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the large zone high grey level emphasis

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    r_vect = range(1, sz[0]+1)  # Column vectors
    # Column and row indicators for each entry of the glszm
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)

    # Large zone high grey level emphasis
    return np.sum(np.sum(glszm*(np.power(r_mat, 2))*(np.power(c_mat, 2))))

def glnu(glszm: np.ndarray) -> float:
    """Computes grey level non-uniformity feature.
    This feature refers to "Fszm_glnu" (ID = JNSA) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the grey level non-uniformity feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    n_s = np.sum(glszm)

    pg = np.transpose(np.sum(glszm, 1))  # Gray-Level Vector

    # Grey level non-uniformity feature
    return np.sum(np.power(pg, 2)) * n_s

def glnu_norm(glszm: np.ndarray) -> float:
    """Computes grey level non-uniformity normalised 
    This feature refers to "Fszm_glnu_norm" (ID = Y1RO) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the grey level non-uniformity normalised feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm

    pg = np.transpose(np.sum(glszm, 1))  # Gray-Level Vector

    # Grey level non-uniformity normalised feature
    return np.sum(np.power(pg, 2))

def zsnu(glszm: np.ndarray) -> float:
    """Computes zone size non-uniformity 
    This feature refers to "Fszm_zsnu" (ID = 4JP3) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the zone size non-uniformity feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    n_s = np.sum(glszm)

    pz = np.sum(glszm, 0)  # Zone Size Vector

    # Zone size non-uniformity feature
    return np.sum(np.power(pz, 2)) * n_s

def zsnu_norm(glszm: np.ndarray) -> float:
    """Computes zone size non-uniformity normalised 
    This feature refers to "Fszm_zsnu_norm" (ID = VB3A) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the zone size non-uniformity normalised feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm

    pz = np.sum(glszm, 0)  # Zone Size Vector

    # Zone size non-uniformity normalised feature
    return np.sum(np.power(pz, 2))

def z_perc(glszm: np.ndarray) -> float:
    """Computes zone percentage 
    This feature refers to "Fszm_z_perc" (ID = P30P) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the zone percentage feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    pg = np.transpose(np.sum(glszm, 1))  # Gray-Level Vector
    pz = np.sum(glszm, 0)  # Zone Size Vector

    # Zone percentage feature
    return np.sum(pg)/(np.matmul(pz, np.transpose(c_vect)))

def gl_var(glszm: np.ndarray) -> float:
    """Computes grey level variance 
    This feature refers to "Fszm_gl_var" (ID = BYLV) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the grey level variance feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    r_vect = range(1, sz[0]+1)  # Column vectors
    # Column and row indicators for each entry of the glszm
    _, r_mat = np.meshgrid(c_vect, r_vect)

    temp = r_mat * glszm
    u = np.sum(temp)
    temp = (np.power(r_mat - u, 2)) * glszm

    # Grey level variance feature
    return np.sum(temp)

def zs_var(glszm: np.ndarray) -> float:
    """Computes zone size variance 
    This feature refers to "Fszm_zs_var" (ID = 3NSA) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the zone size variance feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm
    sz = np.shape(glszm)  # Size of glszm

    c_vect = range(1, sz[1]+1)  # Row vectors
    r_vect = range(1, sz[0]+1)  # Column vectors
    # Column and row indicators for each entry of the glszm
    c_mat, _ = np.meshgrid(c_vect, r_vect)

    temp = c_mat * glszm
    u = np.sum(temp)
    temp = (np.power(c_mat - u, 2)) * glszm

    # Zone size variance feature
    return np.sum(temp)

def zs_entr(glszm: np.ndarray) -> float:
    """Computes zone size entropy 
    This feature refers to "Fszm_zs_entr" (ID = GU8N) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        glszm (ndarray): array of the gray level size zone matrix
    
    Returns:
        float: the zone size entropy feature

    """
    glszm = glszm/np.sum(glszm)  # Normalization of glszm

    val_pos = glszm[np.nonzero(glszm)]
    temp = val_pos * np.log2(val_pos)

    # Zone size entropy feature
    return -np.sum(temp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import warnings
from typing import Dict, Tuple

import numpy as np
from scipy.stats import scoreatpercentile, variation


def init_IH(vol: np.ndarray) -> Tuple[np.ndarray, np.ndarray, int, np.ndarray, np.ndarray]:
    """Initialize Intensity Histogram Features.

    Args:
        vol (ndarray): 3D volume, QUANTIZED (e.g. nBins = 100, 
            levels = [1, ..., max]), with NaNs outside the region of interest.

    Returns:
        Dict: Dict of the Intensity Histogram Features.
    """
    warnings.simplefilter("ignore")

    # INITIALIZATION

    x = vol[~np.isnan(vol[:])]
    n_v = x.size

    # CONSTRUCTION OF HISTOGRAM AND ASSOCIATED NUMBER OF GRAY-LEVELS

    # Always defined from 1 to the maximum value of
    # the volume to remove any ambiguity
    levels = np.arange(1, np.max(x) + 100*np.finfo(float).eps)
    n_g = levels.size  # Number of gray-levels
    h = np.zeros(n_g)  # The histogram of x

    for i in np.arange(0, n_g):
        # == i or == levels(i) is equivalent since levels = 1:max(x),
        # and n_g = numel(levels)
        h[i] = np.sum(x == i + 1)  # h[i] = sum(x == i+1)

    p = (h / n_v)  # Occurence probability for each grey level bin i
    pt = p.transpose()

    return x, levels, n_g, h, p, pt

def extract_all(vol: np.ndarray) -> Dict:
    """Computes Intensity Histogram Features.
    These features refer to Intensity histogram family in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, QUANTIZED (e.g. nBins = 100,
                       levels = [1, ..., max]), with NaNs outside the region of interest.

    Returns:
        Dict: Dict of the Intensity Histogram Features.
    """
    warnings.simplefilter("ignore")

    # INITIALIZATION
    x, levels, n_g, h, p, pt = init_IH(vol)

    # Initialization of final structure (Dictionary) containing all features.
    int_hist = {'Fih_mean': [],
               'Fih_var': [],
               'Fih_skew': [],
               'Fih_kurt': [],
               'Fih_median': [],
               'Fih_min': [],
               'Fih_P10': [],
               'Fih_P90': [],
               'Fih_max': [],
               'Fih_mode': [],
               'Fih_iqr': [],
               'Fih_range': [],
               'Fih_mad': [],
               'Fih_rmad': [],
               'Fih_medad': [],
               'Fih_cov': [],
               'Fih_qcod': [],
               'Fih_entropy': [],
               'Fih_uniformity': [],
               'Fih_max_grad': [],
               'Fih_max_grad_gl': [],
               'Fih_min_grad': [],
               'Fih_min_grad_gl': []
               }

    # STARTING COMPUTATION
    # Intensity histogram mean
    u = np.matmul(levels, pt)
    int_hist['Fih_mean'] = u

    # Intensity histogram variance
    var = np.matmul(np.power(levels - u, 2), pt)
    int_hist['Fih_var'] = var

    # Intensity histogram skewness and kurtosis
    skew = 0
    kurt = 0

    if var != 0:
        skew = np.matmul(np.power(levels - u, 3), pt) / np.power(var, 3/2)
        kurt = np.matmul(np.power(levels - u, 4), pt) / np.power(var, 2) - 3

    int_hist['Fih_skew'] = skew
    int_hist['Fih_kurt'] = kurt

    # Intensity histogram median
    int_hist['Fih_median'] = np.median(x)

    # Intensity histogram minimum grey level
    int_hist['Fih_min'] = np.min(x)

    # Intensity histogram 10th percentile
    p10 = scoreatpercentile(x, 10)
    int_hist['Fih_P10'] = p10

    # Intensity histogram 90th percentile
    p90 = scoreatpercentile(x, 90)
    int_hist['Fih_P90'] = p90

    # Intensity histogram maximum grey level
    int_hist['Fih_max'] = np.max(x)

    # Intensity histogram mode
    #    levels = 1:max(x), so the index of the ith bin of h is the same as i
    mh = np.max(h)
    mode = np.where(h == mh)[0] + 1

    if np.size(mode) > 1:
        dist = np.abs(mode - u)
        ind_min = np.argmin(dist)
        int_hist['Fih_mode'] = mode[ind_min]
    else:
        int_hist['Fih_mode'] = mode[0]

    # Intensity histogram interquantile range
    #    Since x goes from 1:max(x), all with integer values,
    #    the result is an integer
    int_hist['Fih_iqr'] = scoreatpercentile(x, 75) - scoreatpercentile(x, 25)

    # Intensity histogram range
    int_hist['Fih_range'] = np.max(x) - np.min(x)

    # Intensity histogram mean absolute deviation
    int_hist['Fih_mad'] = np.mean(abs(x - u))

    # Intensity histogram robust mean absolute deviation
    x_10_90 = x[np.where((x >= p10) & (x <= p90), True, False)]
    int_hist['Fih_rmad'] = np.mean(np.abs(x_10_90 - np.mean(x_10_90)))

    # Intensity histogram median absolute deviation
    int_hist['Fih_medad'] = np.mean(np.absolute(x - np.median(x)))

    # Intensity histogram coefficient of variation
    int_hist['Fih_cov'] = variation(x)

    # Intensity histogram quartile coefficient of dispersion
    x_75_25 = scoreatpercentile(x, 75) + scoreatpercentile(x, 25)
    int_hist['Fih_qcod'] = int_hist['Fih_iqr'] / x_75_25

    # Intensity histogram entropy
    p = p[p > 0]
    int_hist['Fih_entropy'] = -np.sum(p * np.log2(p))

    # Intensity histogram uniformity
    int_hist['Fih_uniformity'] = np.sum(np.power(p, 2))

    # Calculation of histogram gradient
    hist_grad = np.zeros(n_g)
    hist_grad[0] = h[1] - h[0]
    hist_grad[-1] = h[-1] - h[-2]

    for i in np.arange(1, n_g-1):
        hist_grad[i] = (h[i+1] - h[i-1])/2

    # Maximum histogram gradient
    int_hist['Fih_max_grad'] = np.max(hist_grad)

    # Maximum histogram gradient grey level
    ind_max = np.where(hist_grad == int_hist['Fih_max_grad'])[0][0]
    int_hist['Fih_max_grad_gl'] = levels[ind_max]

    # Minimum histogram gradient
    int_hist['Fih_min_grad'] = np.min(hist_grad)

    # Minimum histogram gradient grey level
    ind_min = np.where(hist_grad == int_hist['Fih_min_grad'])[0][0]
    int_hist['Fih_min_grad_gl'] = levels[ind_min]

    return int_hist

def mean(vol: np.ndarray) -> float:
    """Compute Intensity histogram mean feature of the input dataset (3D Array).
    This feature refers to "Fih_mean" (ID = X6K6) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram mean
    """
    _, levels, _, _, _, pt = init_IH(vol)  # Initialization
    
    return np.matmul(levels, pt) # Intensity histogram mean

def var(vol: np.ndarray) -> float:
    """Compute Intensity histogram variance feature of the input dataset (3D Array).
    This feature refers to "Fih_var" (ID = CH89) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram variance
    """
    _, levels, _, _, _, pt = init_IH(vol)  # Initialization
    u = np.matmul(levels, pt) # Intensity histogram mean

    return np.matmul(np.power(levels - u, 2), pt)  # Intensity histogram variance

def skewness(vol: np.ndarray) -> float:
    """Compute Intensity histogram skewness feature of the input dataset (3D Array).
    This feature refers to "Fih_skew" (ID = 88K1) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram skewness.
    """
    _, levels, _, _, _, pt = init_IH(vol)  # Initialization
    u = np.matmul(levels, pt) # Intensity histogram mean
    var = np.matmul(np.power(levels - u, 2), pt)  # Intensity histogram variance
    if var != 0:
        skew = np.matmul(np.power(levels - u, 3), pt) / np.power(var, 3/2)

    return skew  # Skewness

def kurt(vol: np.ndarray) -> float:
    """Compute Intensity histogram kurtosis feature of the input dataset (3D Array).
    This feature refers to "Fih_kurt" (ID = C3I7) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: The Intensity histogram kurtosis feature
    """
    _, levels, _, _, _, pt = init_IH(vol)  # Initialization
    u = np.matmul(levels, pt) # Intensity histogram mean
    var = np.matmul(np.power(levels - u, 2), pt)  # Intensity histogram variance
    if var != 0:
        kurt = np.matmul(np.power(levels - u, 4), pt) / np.power(var, 2) - 3

    return kurt  # Kurtosis

def median(vol: np.ndarray) -> float:
    """Compute Intensity histogram median feature along the specified axis of the input dataset (3D Array).
    This feature refers to "Fih_median" (ID = WIFQ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram median feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.median(x)  # Median

def min(vol: np.ndarray) -> float:
    """Compute Intensity histogram minimum grey level feature of the input dataset (3D Array).
    This feature refers to "Fih_min" (ID = 1PR8) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram minimum grey level feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.min(x)  # Minimum grey level

def p10(vol: np.ndarray) -> float:
    """Compute Intensity histogram 10th percentile feature of the input dataset (3D Array).
    This feature refers to "Fih_P10" (ID = GPMT) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram 10th percentile feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return scoreatpercentile(x, 10)  # 10th percentile

def p90(vol: np.ndarray) -> float:
    """Compute Intensity histogram 90th percentile feature of the input dataset (3D Array).
    This feature refers to "Fih_P90" (ID = OZ0C) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    Returns:
        float: Intensity histogram 90th percentile feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return scoreatpercentile(x, 90)  # 90th percentile

def max(vol: np.ndarray) -> float:
    """Compute Intensity histogram maximum grey level feature of the input dataset (3D Array).
    This feature refers to "Fih_max" (ID = 3NCY) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram maximum grey level feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.max(x)  # Maximum grey level

def mode(vol: np.ndarray) -> int:
    """Compute Intensity histogram mode feature of the input dataset (3D Array).
    This feature refers to "Fih_mode" (ID = AMMC) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        integer: Intensity histogram mode. 
        levels = 1:max(x), so the index of the ith bin of h is the same as i
    """
    _, levels, _, h, _, pt = init_IH(vol)  # Initialization
    u = np.matmul(levels, pt)
    mh = np.max(h)
    mode = np.where(h == mh)[0] + 1

    if np.size(mode) > 1:
        dist = np.abs(mode - u)
        ind_min = np.argmin(dist)

        return mode[ind_min]  # Intensity histogram mode.
    else:

        return mode[0]  # Intensity histogram mode.

def iqrange(vol: np.ndarray) -> float:
    r"""Compute Intensity histogram interquantile range feature of the input dataset (3D Array).
    This feature refers to "Fih_iqr" (ID = WR0O) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Interquartile range. If :math:`axis ≠ None` , the output data-type is the same as that of the input.
        Since x goes from :math:`1:max(x)` , all with integer values, the result is an integer.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return scoreatpercentile(x, 75) - scoreatpercentile(x, 25)  # Intensity histogram interquantile range

def range(vol: np.ndarray) -> float:
    """Compute Intensity histogram range of values (maximum - minimum) feature of the input dataset (3D Array).
    This feature refers to "Fih_range" (ID = 5Z3W) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram range.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.max(x) - np.min(x) # Intensity histogram range

def mad(vol: np.ndarray) -> float:
    """Compute Intensity histogram mean absolute deviation feature of the input dataset (3D Array).
    This feature refers to "Fih_mad" (ID = D2ZX) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float : Intensity histogram mean absolute deviation feature.
    """
    x, levels, _, _, _, pt = init_IH(vol)  # Initialization
    u = np.matmul(levels, pt)

    return np.mean(abs(x - u))  # Intensity histogram mean absolute deviation

def rmad(vol: np.ndarray) -> float:
    """Compute Intensity histogram robust mean absolute deviation feature of the input dataset (3D Array).
    This feature refers to "Fih_rmad" (ID = WRZB) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
        P10(ndarray): Score at 10th percentil.
        P90(ndarray): Score at 90th percentil.
    
    Returns:
        float: Intensity histogram robust mean absolute deviation
    """
    x = vol[~np.isnan(vol[:])]  # Initialization
    P10 = scoreatpercentile(x, 10)  # 10th percentile
    P90 = scoreatpercentile(x, 90)  # 90th percentile
    x_10_90 = x[np.where((x >= P10) &
                         (x <= P90), True, False)]  # Holding x for (x >= P10) and (x<= P90)

    return np.mean(np.abs(x_10_90 - np.mean(x_10_90)))  # Intensity histogram robust mean absolute deviation

def medad(vol: np.ndarray) -> float:
    """Intensity histogram median absolute deviation feature of the input dataset (3D Array).
    This feature refers to "Fih_medad" (ID = 4RNL) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram median absolute deviation feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return np.mean(np.absolute(x - np.median(x)))  # Intensity histogram median absolute deviation

def cov(vol: np.ndarray) -> float:
    """Compute Intensity histogram coefficient of variation feature of the input dataset (3D Array).
    This feature refers to "Fih_cov" (ID = CWYJ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram coefficient of variation feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization

    return variation(x)  # Intensity histogram coefficient of variation

def qcod(vol: np.ndarray) -> float:
    """Compute the quartile coefficient of dispersion feature of the input dataset (3D Array).
    This feature refers to "Fih_qcod" (ID = SLWD) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        ndarray: A new array holding the quartile coefficient of dispersion feature.
    """
    x = vol[~np.isnan(vol[:])]  # Initialization
    x_75_25 = scoreatpercentile(x, 75) + scoreatpercentile(x, 25)     

    return iqrange(x) / x_75_25  # Quartile coefficient of dispersion

def entropy(vol: np.ndarray) -> float:
    """Compute Intensity histogram entropy feature of the input dataset (3D Array).
    This feature refers to "Fih_entropy" (ID = TLU2) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram entropy feature.
    """
    x, _, _, _, p, _ = init_IH(vol)  # Initialization
    p = p[p > 0]

    return -np.sum(p * np.log2(p))  # Intensity histogram entropy

def uniformity(vol: np.ndarray) -> float:
    """Compute Intensity histogram uniformity feature of the input dataset (3D Array).
    This feature refers to "Fih_uniformity" (ID = BJ5W) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Intensity histogram uniformity feature.
    """
    x, _, _, _, p, _ = init_IH(vol)  # Initialization
    p = p[p > 0]

    return np.sum(np.power(p, 2))  # Intensity histogram uniformity

def hist_grad_calc(vol: np.ndarray) -> np.ndarray:
    """Calculation of histogram gradient.
    This feature refers to "Fih_hist_grad_calc" (ID = 12CE) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        ndarray: Histogram gradient
    """
    _, _, n_g, h, _, _ = init_IH(vol)  # Initialization
    hist_grad = np.zeros(n_g)
    hist_grad[0] = h[1] - h[0]
    hist_grad[-1] = h[-1] - h[-2]
    for i in np.arange(1, n_g-1):
        hist_grad[i] = (h[i+1] - h[i-1])/2

    return hist_grad  # Intensity histogram uniformity

def max_grad(vol: np.ndarray) -> float:
    """Calculation of Maximum histogram gradient feature.
    This feature refers to "Fih_max_grad" (ID = 12CE) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Maximum histogram gradient feature.
    """
    hist_grad = hist_grad_calc(vol)  # Initialization

    return np.max(hist_grad)  # Maximum histogram gradient

def max_grad_gl(vol: np.ndarray) -> float:
    """Calculation of Maximum histogram gradient grey level feature.
    This feature refers to "Fih_max_grad_gl" (ID = 8E6O) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Maximum histogram gradient grey level feature.
    """
    _, levels, _, _, _, _ = init_IH(vol)  # Initialization
    hist_grad = hist_grad_calc(vol) 
    ind_max = np.where(hist_grad == np.max(hist_grad))[0][0]  

    return levels[ind_max]  # Maximum histogram gradient grey level

def min_grad(vol: np.ndarray) -> float:
    """Calculation of Minimum histogram gradient feature.
    This feature refers to "Fih_min_grad" (ID = VQB3) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Minimum histogram gradient feature.
    """
    hist_grad = hist_grad_calc(vol)  # Initialization

    return np.min(hist_grad)  # Minimum histogram gradient

def min_grad_gl(vol: np.ndarray) -> float:
    """Calculation of Minimum histogram gradient grey level feature.
    This feature refers to "Fih_min_grad_gl" (ID = RHQZ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol(ndarray): 3D volume, NON-QUANTIZED, with NaNs outside the region of interest
            (continuous imaging intensity distribution)
    
    Returns:
        float: Minimum histogram gradient grey level feature.
    """
    _, levels, _, _, _, _ = init_IH(vol)  # Initialization
    hist_grad = hist_grad_calc(vol) 
    ind_min = np.where(hist_grad == np.min(hist_grad))[0][0] 

    return levels[ind_min]  # Minimum histogram gradient grey level



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Dict

from numpy import ndarray

from ..biomarkers.utils import get_glob_peak, get_loc_peak


def extract_all(img_obj: ndarray,
                roi_obj: ndarray,
                res: ndarray,
                intensity_type: str,
                compute_global: bool = False) -> Dict:
    """Compute Local Intensity Features.
    This features refer to Local Intensity family in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        img_obj (ndarray): Continuous image intensity distribution, with no NaNs
            outside the ROI.
        roi_obj (ndarray): Array of the mask defining the ROI.
        res (List[float]): [a,b,c] vector specifying the resolution of the volume in mm.
            XYZ resolution (world), or JIK resolution (intrinsic matlab).
        intensity_type (str): Type of intensity to compute. Can be "arbitrary", "definite" or "filtered".
            Will compute features only for "definite" intensity type.
        compute_global (bool, optional): If True, will compute global intensity peak, we
            recommend you don't set it to True if not necessary in your study or analysis as it
            takes too much time for calculation. Default: False.

    Returns:
        Dict: Dict of the Local Intensity Features.

    Raises:
        ValueError: If `intensity_type` is not "arbitrary", "definite" or "filtered".
    """
    assert intensity_type in ["arbitrary", "definite", "filtered"], \
        "intensity_type must be 'arbitrary', 'definite' or 'filtered'"
    
    loc_int = {'Floc_peak_local': [], 'Floc_peak_global': []}

    if intensity_type == "definite":
        loc_int['Floc_peak_local'] = (get_loc_peak(img_obj, roi_obj, res))

    # NEEDS TO BE VECTORIZED FOR FASTER CALCULATION! OR
    # SIMPLY JUST CONVOLUTE A 3D AVERAGING FILTER!
    if compute_global:
        loc_int['Floc_peak_global'] = (get_glob_peak(img_obj,roi_obj, res))

    return loc_int

def peak_local(img_obj: ndarray,
               roi_obj: ndarray,
               res: ndarray) -> float:
    """Computes local intensity peak.
    This feature refers to "Floc_peak_local" (ID = VJGA) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        img_obj (ndarray): Continuous image intensity distribution, with no NaNs
            outside the ROI.
        roi_obj (ndarray): Array of the mask defining the ROI.
        res (List[float]): [a,b,c] vector specifying the resolution of the volume in mm.
            XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Local intensity peak.
    """
    return get_loc_peak(img_obj, roi_obj, res)

def peak_global(img_obj: ndarray,
                roi_obj: ndarray,
                res: ndarray) -> float:
    """Computes global intensity peak.
    This feature refers to "Floc_peak_global" (ID = 0F91) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        img_obj (ndarray): Continuous image intensity distribution, with no NaNs
            outside the ROI.
        roi_obj (ndarray): Array of the mask defining the ROI.
        res (List[float]): [a,b,c] vector specifying the resolution of the volume in mm.
            XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Global intensity peak.
    """
    return get_glob_peak(img_obj, roi_obj, res)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List, Union

import numpy as np


def get_neighbour_direction(d=1.8,
                            distance="euclidian",
                            centre=False,
                            complete=False,
                            dim3=True) -> np.ndarray:
    """Defines transitions to neighbour voxels.

    Note:
        This code was adapted from the in-house radiomics software created at
        OncoRay, Dresden, Germany.

    Args:
        d (float, optional): Max ``distance`` between voxels.
        distance (str, optional): Distance norm used to compute distances. must be
                                  "manhattan", "l1", "l_1", "euclidian", "l2", "l_2", "chebyshev", "linf" or "l_inf".
        centre (bool, optional): Flags whether the [0,0,0] direction should be included
        complete(bool, optional): Flags whether all directions should be computed (True)
                                  or just the primary ones (False). For example, including [0,0,1] and [0,0,-1]
                                  directions may lead to redundant texture matrices.
        dim3(bool, optional): flags whether full 3D (True) or only in-slice (2D; False)
                              directions should be considered.

    Returns:
        ndarray: set of k neighbour direction vectors.
    """

    # Base transition vector
    trans = np.arange(start=-np.ceil(d), stop=np.ceil(d)+1)
    n = np.size(trans)

    # Build transition array [x,y,z]
    nbrs = np.array([rep(x=trans, each=n * n, times=1),
                     rep(x=trans, each=n, times=n),
                     rep(x=trans, each=1, times=n * n)], dtype=np.int32)

    # Initiate maintenance index
    index = np.zeros(np.shape(nbrs)[1], dtype=bool)

    # Remove neighbours more than distance d from the center ----------------

    # Manhattan distance
    if distance.lower() in ["manhattan", "l1", "l_1"]:
        index = np.logical_or(index, np.sum(np.abs(nbrs), axis=0) <= d)
    # Eucldian distance
    if distance.lower() in ["euclidian", "l2", "l_2"]:
        index = np.logical_or(index, np.sqrt(
            np.sum(np.multiply(nbrs, nbrs), axis=0)) <= d)
    # Chebyshev distance
    if distance.lower() in ["chebyshev", "linf", "l_inf"]:
        index = np.logical_or(index, np.max(np.abs(nbrs), axis=0) <= d)

    # Check if centre voxel [0,0,0] should be maintained; False indicates removal
    if centre is False:
        index = np.logical_and(index, (np.sum(np.abs(nbrs), axis=0)) > 0)

    # Check if a complete neighbourhood should be returned
    # False indicates that only half of the vectors are returned
    if complete is False:
        index[np.arange(start=0, stop=len(index)//2 + 1)] = False

    # Check if neighbourhood should be 3D or 2D
    if dim3 is False:
        index[nbrs[2, :] != 0] = False

    return nbrs[:, index]


def rep(x: np.ndarray,
        each=1,
        times=1) -> np.ndarray:
    """Replicates the values in ``x``.
    Replicates the :func:`"rep"` function found in R for tiling and repeating vectors.

    Note:
        Code was adapted from the in-house radiomics software created at OncoRay,
        Dresden, Germany.

    Args:
        x (ndarray): Array to replicate.
        each (int): Integer (non-negative) giving the number of times to repeat
                    each element of the passed array.
        times (int): Integer (non-negative). Each element of ``x`` is repeated each times.

    Returns:
        ndarray: Array with same values but replicated.
    """

    each = int(each)
    times = int(times)

    if each > 1:
        x = np.repeat(x, repeats=each)

    if times > 1:
        x = np.tile(x, reps=times)

    return x

def get_value(x: np.ndarray,
              index: int,
              replace_invalid=True) -> np.ndarray:
    """Retrieves intensity values from an image intensity table used for computing
    texture features.

    Note:
        Code was adapted from the in-house radiomics software created at OncoRay,
        Dresden, Germany.

    Args:
        x (ndarray): set of intensity values.
        index (int): Index to the provided set of intensity values.
        replace_invalid (bool, optional): If True, invalid indices will be replaced
                                          by a placeholder "NaN" value.

    Returns:
        ndarray: Array of the intensity values found at the requested indices.

    """

    # Initialise placeholder
    read_x = np.zeros(np.shape(x))

    # Read variables for valid indices
    read_x[index >= 0] = x[index[index >= 0]]

    if replace_invalid:
        # Set variables for invalid indices to nan
        read_x[index < 0] = np.nan

        # Set variables for invalid initial indices to nan
        read_x[np.isnan(x)] = np.nan

    return read_x


def coord2index(x: np.ndarray,
                y: np.ndarray,
                z: np.ndarray,
                dims: Union[List, np.ndarray]) -> Union[np.ndarray,
                                                        List]:
    """Translate requested coordinates to row indices in image intensity tables.

    Note:
        Code was adapted from the in-house radiomics software created at OncoRay,
        Dresden, Germany.

    Args:
        x (ndarray): set of discrete x-coordinates.
        y (ndarray): set of discrete y-coordinates.
        z (ndarray): set of discrete z-coordinates.
        dims (ndarray or List): dimensions of the image.

    Returns:
        ndarray or List: Array or List of indexes corresponding the requested coordinates

    """

    # Translate coordinates to indices
    index = z + y * dims[2] + x * dims[2] * dims[1]

    # Mark invalid transitions
    index[np.logical_or(x < 0, x >= dims[0])] = -99999
    index[np.logical_or(y < 0, y >= dims[1])] = -99999
    index[np.logical_or(z < 0, z >= dims[2])] = -99999

    return index


def is_list_all_none(x: List) -> bool:
    """Determines if all list elements are None.

    Args:
        x (List): List of elements to check.

    Returns:
        bool: True if all elemets in `x` are None.

    """
    return all(y is None for y in x)



import math
from itertools import product
from typing import List, Union

import numpy as np

from ..MEDscan import MEDscan
from ..utils.image_volume_obj import image_volume_obj
from .utils import convolve


class Gabor():
    """
    The Gabor filter class
    """

    def __init__(
            self,
            size: int,
            sigma: float,
            lamb: float,
            gamma: float,
            theta: float,
            rot_invariance=False,
            padding="symmetric"
        ) -> None:
        """
        The constructor of the Gabor filter. Highly inspired by Ref 1.

        Args:
            size (int): An integer that represent the length along one dimension of the kernel.
            sigma (float): A positive float that represent the scale of the Gabor filter
            lamb (float): A positive float that represent the wavelength in the Gabor filter. (mm or pixel?)
            gamma (float): A positive float that represent the spacial aspect ratio
            theta (float): Angle parameter used in the rotation matrix
            rot_invariance (bool): If true, rotation invariance will be done on the kernel and the kernel
                                   will be rotate 2*pi / theta times.
            padding: The padding type that will be used to produce the convolution

        Returns:
            None
        """

        assert ((size + 1) / 2).is_integer() and size > 0, "size should be a positive odd number."
        assert sigma > 0, "sigma should be a positive float"
        assert lamb > 0, "lamb represent the wavelength, so it should be a positive float"
        assert gamma > 0, "gamma is the ellipticity of the support of the filter, so it should be a positive float"

        self.dim = 2
        self.padding = padding
        self.size = size
        self.sigma = sigma
        self.lamb = lamb
        self.gamma = gamma
        self.theta = theta
        self.rot = rot_invariance
        self.create_kernel()

    def create_kernel(self) -> List[np.ndarray]:
        """Create the kernel of the Gabor filter
    
        Returns:
            List[ndarray]: A list of numpy 2D-array that contain the kernel of the real part and
            the imaginary part respectively.
        """

        def compute_weight(position, theta):
            k_2 = position[0]*math.cos(theta) + position[1] * math.sin(theta)
            k_1 = position[1]*math.cos(theta) - position[0] * math.sin(theta)

            common = math.e**(-(k_1**2 + (self.gamma*k_2)**2)/(2*self.sigma**2))
            real = math.cos(2*math.pi*k_1/self.lamb)
            im = math.sin(2*math.pi*k_1/self.lamb)
            return common*real, common*im

        # Rotation invariance
        nb_rot = round(2*math.pi/abs(self.theta)) if self.rot else 1
        real_list = []
        im_list = []

        for i in range(1, nb_rot+1):
            # Initialize the kernel as tensor of zeros
            real_kernel = np.zeros([self.size for _ in range(2)])
            im_kernel = np.zeros([self.size for _ in range(2)])

            for k in product(range(self.size), repeat=2):
                real_kernel[k], im_kernel[k] = compute_weight(np.array(k)-int((self.size-1)/2), self.theta*i)

            real_list.extend([real_kernel])
            im_list.extend([im_kernel])

        self.kernel = np.expand_dims(
            np.concatenate((real_list, im_list), axis=0),
            axis=1
        )

    def convolve(self,
                 images: np.ndarray,
                 orthogonal_rot=False,
                 pooling_method='mean') -> np.ndarray:
        """Filter a given image using the Gabor kernel defined during the construction of this instance.

        Args:
            images (ndarray): A n-dimensional numpy array that represent the images to filter
            orthogonal_rot (bool): If true, the 3D images will be rotated over coronal, axial and sagittal axis

        Returns:
            ndarray: The filtered image as a numpy ndarray
        """

        # Swap the second axis with the last, to convert image B, W, H, D --> B, D, H, W
        image = np.swapaxes(images, 1, 3)

        result = convolve(self.dim, self.kernel, image, orthogonal_rot, self.padding)

        # Reshape to get real and imaginary response on the first axis.
        _dim = 2 if orthogonal_rot else 1
        nb_rot = int(result.shape[_dim]/2)
        result = np.stack(np.array_split(result, np.array([nb_rot]), _dim), axis=0)

        # 2D modulus response map
        result = np.linalg.norm(result, axis=0)

        # Rotation invariance.
        if pooling_method == 'mean':
            result = np.mean(result, axis=2) if orthogonal_rot else np.mean(result, axis=1)
        elif pooling_method == 'max':
            result = np.max(result, axis=2) if orthogonal_rot else np.max(result, axis=1)
        else:
            raise ValueError("Pooling method should be either 'mean' or 'max'.")

        # Aggregate orthogonal rotation
        result = np.mean(result, axis=0) if orthogonal_rot else result
            
        return np.swapaxes(result, 1, 3)

def apply_gabor(
        input_images: Union[image_volume_obj, np.ndarray],
        medscan: MEDscan = None,
        voxel_length: float = 0.0,
        sigma: float = 0.0,
        _lambda: float = 0.0,
        gamma: float = 0.0,
        theta: float = 0.0,
        rot_invariance: bool = False,
        padding: str = "symmetric",
        orthogonal_rot: bool = False,
        pooling_method: str = "mean"
    ) -> np.ndarray:
    """Apply the Gabor filter to a given imaging data.
    
    Args:
        input_images (Union[image_volume_obj, np.ndarray]): The input images to filter.
        medscan (MEDscan, optional): The MEDscan object that will provide the filter parameters.
        voxel_length (float, optional): The voxel size of the input image.
        sigma (float, optional): A positive float that represent the scale of the Gabor filter.
        _lambda (float, optional): A positive float that represent the wavelength in the Gabor filter.
        gamma (float, optional): A positive float that represent the spacial aspect ratio.
        theta (float, optional): Angle parameter used in the rotation matrix.
        rot_invariance (bool, optional): If true, rotation invariance will be done on the kernel and the kernel
            will be rotate 2*pi / theta times.
        padding (str, optional): The padding type that will be used to produce the convolution. Check options 
            here: `numpy.pad <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`__.
        orthogonal_rot (bool, optional): If true, the 3D images will be rotated over coronal, axial and sagittal axis.
    
    Returns:
        ndarray: The filtered image.
    """
    # Check if the input is a numpy array or a Image volume object
    spatial_ref = None
    if type(input_images) == image_volume_obj:
        spatial_ref = input_images.spatialRef
        input_images = input_images.data
    
    # Convert to shape : (B, W, H, D)
    input_images = np.expand_dims(input_images.astype(np.float64), axis=0) 

    if medscan:
        # Initialize filter class params & instance
        voxel_length = medscan.params.process.scale_non_text[0]
        sigma = medscan.params.filter.gabor.sigma / voxel_length
        lamb = medscan.params.filter.gabor._lambda / voxel_length
        size = 2 * int(7 * sigma + 0.5) + 1
        _filter = Gabor(size=size,
                        sigma=sigma,
                        lamb=lamb,
                        gamma=medscan.params.filter.gabor.gamma,
                        theta=-medscan.params.filter.gabor.theta,
                        rot_invariance=medscan.params.filter.gabor.rot_invariance,
                        padding=medscan.params.filter.gabor.padding
                        )
        # Run convolution
        result = _filter.convolve(input_images, orthogonal_rot=medscan.params.filter.gabor.orthogonal_rot)
    else:
        if not (voxel_length and sigma and _lambda and gamma and theta):
            raise ValueError("Missing parameters to build the Gabor filter.")
        # Initialize filter class params & instance
        sigma = sigma / voxel_length
        lamb = _lambda / voxel_length
        size = 2 * int(7 * sigma + 0.5) + 1
        _filter = Gabor(size=size,
                        sigma=sigma,
                        lamb=lamb,
                        gamma=gamma,
                        theta=theta,
                        rot_invariance=rot_invariance,
                        padding=padding
                        )
        # Run convolution
        result = _filter.convolve(input_images, orthogonal_rot=orthogonal_rot, pooling_method=pooling_method)
    
    if spatial_ref:
        return image_volume_obj(np.squeeze(result), spatial_ref)
    else:
        return np.squeeze(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from scipy.ndimage import map_coordinates


def interp3(v, x_q, y_q, z_q, method) -> np.ndarray:
    """`Interpolation for 3-D gridded data <https://www.mathworks.com/help/matlab/ref/interp3.html>`_\
    in meshgrid format, implements similar functionality MATLAB interp3.

    Args:
        X, Y, Z (ndarray) : Query points, should be intrinsic coordinates.
        method (str): {nearest, linear, spline, cubic}, Interpolation ``method``.

    Returns:
        ndarray: Array of interpolated values.
    
    Raises:
        ValueError: If ``method`` is not 'nearest', 'linear', 'spline' or 'cubic'.

    """

    # Parse method
    if method == "nearest":
        spline_order = 0
    elif method == "linear":
        spline_order = 1
    elif method in ["spline", "cubic"]:
        spline_order = 3
    else:
        raise ValueError("Interpolator not implemented.")

    size = np.size(x_q)
    coord_X = np.reshape(x_q, size, order='F')
    coord_Y = np.reshape(y_q, size, order='F')
    coord_Z = np.reshape(z_q, size, order='F')
    coordinates = np.array([coord_X, coord_Y, coord_Z]).astype(np.float32)
    v_q = map_coordinates(input=v.astype(
        np.float32), coordinates=coordinates, order=spline_order, mode='nearest')
    v_q = np.reshape(v_q, np.shape(x_q), order='F')

    return v_q



from typing import List

import numpy as np


def get_full_rad_names(str_user_data: str, rad_var_ids: List):
    """
    Returns the full real names of the radiomics variables (sequential names are not very informative)
    Args:
        str_user_data: string containing the full rad names
        rad_var_ids: can get it by doing table.column.values
    
    Returns:
        List: List of full radiomic names.
    """
    full_rad_names = np.array([])
    for rad_var in rad_var_ids:
        ind_var = int(rad_var[6:])
        full_rad_names = np.append(full_rad_names, str_user_data.split('||')[ind_var].split(':')[1])

    return full_rad_names



import pandas as pd


def get_institutions_from_ids(patient_ids):
    """
    Extracts the institution strings from a cell of patient IDs.

    Args:
        patient_ids (Any): Patient ID (string, list of strings or pandas Series). Ex: 'Cervix-CEM-010'.
    
    Returns:
        str: Categorical vector, specifying the institution of each patient_id entry in "patient_ids". Ex: 'CEM'.
    """
    if isinstance(patient_ids, list):
        patient_ids = pd.Series(patient_ids)
    return patient_ids.str.rsplit('-', expand=True)[1]
